import gdal
import numpy as np
import TronGisPy as tgp
from TronGisPy import Interpolation
from matplotlib import pyplot as plt

class Raster():
    """A Raster object contains all required information for a gis raster file such
    as `.tif` file including digital number for each pixel, number of rows,
    number of cols, number of bands, geo_transform, projection, no_data_value
    and metadata. 

    Examples
    --------
    >>> import TronGisPy as tgp
    >>> raster = tgp.read_raster(tgp.get_testing_fp())
    >>> raster.plot()
    >>> raster
    shape: (677, 674, 3)
    geo_transform: (271982.8783, 1.1186219584569888, 0.0, 2769973.0653, 0.0, -1.1186305760705852)
    projection: PROJCS["TWD97 / TM2 zone 121",GEOGCS["TWD97",DATUM["Taiwan_Datum_1997",SPHEROID["GRS 1980",6378137,298.257222101,AUTHORITY["EPSG","7019"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","1026"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","3824"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",121],PARAMETER["scale_factor",0.9999],PARAMETER["false_easting",250000],PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","3826"]]
    no_data_value: -32768.0
    metadata: {'AREA_OR_POINT': 'Area'}
    >>> raster.data.shape
    (677, 674, 3)
    >>> raster.shape
    (677, 674, 3)
    >>> raster.extent
    array([[ 271982.8783, 2769973.0653],
           [ 272736.8295, 2769973.0653],
           [ 272736.8295, 2769215.7524],
           [ 271982.8783, 2769215.7524]])
    >>> raster.extent_for_plot
    (271982.8783, 272736.8295, 2769215.7524, 2769973.0653)
    """

    def __init__(self, data, geo_transform=None, projection=None, gdaldtype=None, no_data_value=None, metadata=None):
        """Initializing Raster object.

        Parameters
        ----------
        data: array_like
            Digital number for each raster cell. Data is in (n_rows, n_cols, n_bands) shape.
        geo_transform: tuple or list, optional
            Affine transform parameters (c, a, b, f, d, e = geo_transform). 
        projection: str, optional
            The well known text (WKT) of the raster which can be generate 
            from `TronGisPy.epsg_to_wkt(<epsg_code>)`
        gdaldtype: int, optional
            The type of the cell defined in gdal which will affect the information 
            to be stored when saving the file. This can be generate from `gdal.GDT_XXX` 
            such as `gdal.GDT_Int32` equals 5 and `gdal.GDT_Float32` equals 6.
        no_data_value: int or float, optional
            Define which value to replace nan in numpy array when saving a raster file.
        metadata: dict, optional
            Define the metadata of the raster file.
        """
        if len(data.shape) == 2:
            data = np.expand_dims(data, axis=2)
        self.data = data
        self.geo_transform = geo_transform if geo_transform is not None else [0, 1, 0, 0, 0, -1]
        self.gdaldtype = gdaldtype if gdaldtype is not None else tgp.npdtype_to_gdaldtype(data.dtype)
        self.projection = projection
        self.no_data_value = no_data_value
        self.metadata = metadata
        self.cache_data_for_plot = None

    def __repr__(self):
        desc = ""
        desc += "shape: ({rows}, {cols}, {bands})\n".format(rows=self.rows, cols=self.cols, bands=self.bands)
        desc += "gdaldtype: {gdaldtype}\n".format(gdaldtype=tgp.get_gdaldtype_name(self.gdaldtype))
        desc += "geo_transform: {geo_transform}\n".format(geo_transform=self.geo_transform)
        desc += "projection: {projection}\n".format(projection=self.projection)
        desc += "no_data_value: {no_data_value}\n".format(no_data_value=self.no_data_value)
        desc += "metadata: {metadata}".format(metadata=self.metadata)
        return desc

    @property
    def rows(self):
        """Number of rows."""
        return self.data.shape[0]
        
    @property
    def cols(self):
        """Number of cols."""
        return self.data.shape[1]

    @property
    def bands(self):
        """Number of bands."""
        return self.data.shape[2]
        
    @property
    def shape(self):
        """The shape of the raster data."""
        return self.data.shape

    @property
    def gdaldtype_name(self):
        return tgp.get_gdaldtype_name(self.gdaldtype)

    @property
    def data(self):
        """The digital number for each cell of the raster. Data is in (n_rows, n_cols, n_bands) shape."""
        return self.__data

    @data.setter
    def data(self, data):
        """Reset the data"""
        assert type(data) == np.ndarray, "data should be numpy.ndarray type"
        if len(data.shape) == 2:
            data = np.expand_dims(data, axis=2)
        self.__data = data.copy()
        self.update_gdaldtype_by_npdtype()
        self.cache_data_for_plot = None

    @property
    def geo_transform(self):
        """Affine transform parameters (c, a, b, f, d, e = geo_transform)."""
        return self.__geo_transform

    @geo_transform.setter
    def geo_transform(self, geo_transform):
        """Reset the geo_transform"""
        if geo_transform is not None:
            assert len(geo_transform) == 6, "length of geo_transform should be 6"
            self.__geo_transform = tuple(geo_transform)
        else:
            self.__geo_transform = None

    @property
    def metadata(self):
        """Metadata of the raster file."""
        return self.__metadata

    @metadata.setter
    def metadata(self, metadata):
        """Reset the metadata"""
        if metadata is not None:
            assert type(metadata) == dict, "metadata should be in dict type"
            self.__metadata = metadata
        else:
            self.__metadata = None

    @property
    def pixel_size(self):
        """Return pixel resolution on both row and col axes"""
        c, a, b, f, d, e = self.geo_transform 
        res_row = (b**2 + e **2)**(1/2)
        res_col = (a**2 + d**2)**(1/2)
        return (res_row, res_col)

    @property
    def extent(self):
        """Coordinates of Four corner points' of the raster. """
        return tgp.get_extent(self.rows, self.cols, self.geo_transform, return_type='poly')

    @property
    def extent_for_plot(self):
        """(xmin, xmax, ymin, ymax) of the raster boundary."""
        return tgp.get_extent(self.rows, self.cols, self.geo_transform, return_type='plot')

    @property
    def extent_for_gdal(self):
        """(xmin, xmax, ymin, ymax) of the raster boundary."""
        return tgp.get_extent(self.rows, self.cols, self.geo_transform, return_type='gdal')

    # def __getitem__(self, slice_value):
    #     if type(slice_value) in [int, slice]:
    #         h_start_inner, h_stop_inner = self.get_values_by_coords(slice_value)
    #         return self.data[h_start_inner:h_stop_inner, :]

    #     elif type(slice_value) == tuple:
    #         h_start_inner, h_stop_inner = self.get_values_by_coords(slice_value[0])
    #         w_start_inner, w_stop_inner = self.get_values_by_coords(slice_value[1])
    #         return self.data[h_start_inner:h_stop_inner, w_start_inner:w_stop_inner]

    def astype(self, dtype, update_gdaldtype=True):
        """Change dtype of self.data.

        Parameters
        ----------
        dtype: type
            Target dtype.

        update_gdaldtype: bool, optional, default: True
            Change gdaldtype according to `self.data.dtype`.
        """
        assert type(dtype) is type, "dtype should type type"
        self.data = self.data.astype(dtype)
        self.update_gdaldtype_by_npdtype()

    def get_values_by_coords(self, coords):
        """get the data digital values of the coordinates

        Parameters
        ----------
        coords: ndarray 
            The coordinates with shape (n_points, 2). The order of last dimension is (lng, lat).

        Returns
        -------
        coords: ndarray 
            The data digital values of the coordinates.
        """
        npidxs_row, npidxs_col = tgp.coords_to_npidxs(coords.astype(np.float32), self.geo_transform).T
        return self.data[npidxs_row, npidxs_col]

    def update_gdaldtype_by_npdtype(self):
        """Update gdaldtype according to gdaldtype using `TronGisPy.npdtype_to_gdaldtype`.
        For memory operation, numpy dtype will be used. For saving the file,
        gdal dtype will be used. If the data of raster object have been
        changed, its recomended to update the dtype before saveing the file.
        """
        self.gdaldtype = tgp.npdtype_to_gdaldtype(self.data.dtype)

    def to_file(self, fp):
        """Save the file. It is recommended to save the file using tif format, that is 
        use '.tif' as its extension.

        Parameters
        ----------
        fp: str
            File path.
        """
        tgp.write_raster(fp, self.data, self.geo_transform, self.projection, self.gdaldtype, self.no_data_value, self.metadata)

    def to_gdal_ds(self):
        """Export raster object to `gdal.DataSource`.

        Returns
        -------
        ds: gdal.DataSource
            DataSource converted from the raster object.
        """
        ds = tgp.write_gdal_ds(self.data, geo_transform=self.geo_transform, projection=self.projection, 
                                gdaldtype=self.gdaldtype, no_data_value=self.no_data_value)
        return ds

    def fill_na(self, no_data_value=None):
        """Fill np.nan with self.no_data_value

        Parameters
        ----------
        no_data_value: int, optional
            If None, `self.no_data_value` will be used, else self.no_data_value will be re-assigned.
        """
        data = self.data.copy()
        self.no_data_value =  self.no_data_value if no_data_value is None else no_data_value
        data[np.isnan(data)] = self.no_data_value
        self.data = data

    def fill_no_data(self, mode='constant', no_data_value=None, constant=0, window_size=3, loop_to_fill_all=True, loop_limit=5, fill_na=True):
        """Fill no_data cell of the raster object.
        
        Parameters
        ----------
        mode: {'constant', 'neighbor_mean', 'neighbor_majority'}
            the mode used to fill no_data cell.
        no_data_value: int
            If None, `self.no_data_value` will be used, else self.no_data_value will be re-assigned.
        constant: int, optional, default: 0
            If `constant` mode is used, use the constant to fill the cell with values no_data_value.
        window_size: int, optional, default: 3
            If `neighbor_mean` or `neighbor_majority` mode is used, the size of the window of the 
            convolution to calculate the mean value. window_size should be odd number. See also 
            `TronGisPy.Interpolation.mean_interpolation` or `TronGisPy.Interpolation.majority_interpolation`.
        loop_to_fill_all: bool, optional, default: True
            If `neighbor_mean` or `neighbor_majority` mode is used, fill all 
            no_data_value until there is no no_data_value value in the data. See also 
            `TronGisPy.Interpolation.mean_interpolation` or `TronGisPy.Interpolation.majority_interpolation`.
        loop_limit: int, optional, default: 5
            If `neighbor_mean` or `neighbor_majority` mode is used, the maximum 
            limitation on loop. if loop_to_fill_all==True, loop_limit will be considered. `-1` means 
            no limitation. See also `TronGisPy.Interpolation.mean_interpolation` or 
            `TronGisPy.Interpolation.majority_interpolation`.
        """
        self.no_data_value =  self.no_data_value if no_data_value is None else no_data_value
        if fill_na and np.sum(np.isnan(self.data)) > 0:
            self.fill_na(self.no_data_value)

        data = self.data.copy()
        if mode == 'constant':
            data[data == self.no_data_value] = constant
            self.data = data
            self.no_data_value = constant
        elif mode == 'neighbor_mean':
            for i in range(self.bands):
                self.data[:, :, i] = Interpolation.mean_interpolation(data[:, :, i], no_data_value=self.no_data_value, window_size=window_size, loop_to_fill_all=loop_to_fill_all, loop_limit=loop_limit) 
        elif mode == 'neighbor_majority':
            for i in range(self.bands):
                self.data[:, :, i] = Interpolation.majority_interpolation(data[:, :, i], no_data_value=self.no_data_value, window_size=window_size, loop_to_fill_all=loop_to_fill_all, loop_limit=loop_limit)

    def copy(self):
        """copy raster object."""
        return Raster(self.data, self.geo_transform, self.projection, self.gdaldtype, self.no_data_value, self.metadata)

    def refine_resolution(self, dst_resolution, resample_alg='near', extent=None, rotate=True):
        """Refine the resolution of the raster.

        Parameters
        ----------
        dst_resolution: int
            Target Resolution.
        resample_alg: {'near', 'bilinear', 'cubic', 'cubicspline', 'lanczos', 'average', 'mode'}.
            ``near``: nearest neighbour resampling (default, fastest algorithm, worst interpolation quality).
            ``bilinear``: bilinear resampling.
            ``cubic``: cubic resampling.
            ``cubicspline``: cubic spline resampling.
            ``lanczos``: Lanczos windowed sinc resampling.
            ``average``: average resampling, computes the weighted average of all non-NODATA contributing pixels.
            ``mode``: mode resampling, selects the value which appears most often of all the sampled points.
        extent: tuple
            extent to clip the data with (xmin, ymin, xmax, ymax) format.
        rotate: bool
            If True, the function will rotate the raster and adds noData values around it to make a new rectangular image 
            matrix if the rotation in Raster.geo_transform is not zero, else it will keep the original rotation angle of 
            Raster.geo_transform. Gdal will rotate the image by default . Please refer to the issue
            https://gis.stackexchange.com/questions/256081/why-does-gdalwarp-rotate-the-data and 
            https://github.com/OSGeo/gdal/issues/1601.

        Returns
        -------
        dst_raster: Raster
            Refined result.

        Examples
        -------- 
        >>> import numpy as np
        >>> import TronGisPy as tgp
        >>> from TronGisPy import ShapeGrid
        >>> from matplotlib import pyplot as plt
        >>> src_raster_fp = tgp.get_testing_fp('dem_process_path')
        >>> src_raster = tgp.read_raster(src_raster_fp)
        >>> src_raster.data[src_raster.data == -999] = np.nan
        >>> dst_raster = src_raster.refine_resolution(dst_resolution=10, resample_alg='bilinear')
        >>> fig, (ax1, ax2) = plt.subplots(1, 2) # plot the result
        >>> src_raster.plot(ax=ax1)
        >>> ax1.set_title('original dem ' + str(src_raster.shape))
        >>> dst_raster.plot(ax=ax2)
        >>> ax2.set_title('refined image ' + str(dst_raster.shape))
        >>> plt.show()
        """
        src_ds = self.to_gdal_ds()

        if rotate:
            dst_ds = gdal.Warp('', src_ds, xRes=dst_resolution, yRes=dst_resolution, outputBounds=extent, format='MEM', resampleAlg=resample_alg)
            dst_raster = tgp.read_gdal_ds(dst_ds)
        else:
            assert extent is None, "you cannot set the extent when rotate == False"
            zoom_in = dst_resolution / self.pixel_size[0]
            dst_ds = gdal.Warp('', src_ds, xRes=zoom_in, yRes=zoom_in, format='MEM', resampleAlg=resample_alg, transformerOptions=['SRC_METHOD=NO_GEOTRANSFORM', 'DST_METHOD=NO_GEOTRANSFORM'])
            dst_geo_transform = np.array(self.geo_transform)
            dst_geo_transform[[1,2,4,5]] *= zoom_in
            dst_raster = tgp.read_gdal_ds(dst_ds)
            dst_raster.geo_transform = tuple(dst_geo_transform)
            dst_raster.projection = self.projection
        return dst_raster

    def reproject(self, dst_crs='EPSG:4326', src_crs=None):
        """Reproject the raster data.

        Parameters
        ----------
        dst_crs: str, optional, default: EPSG:4326
            The target crs to transform the raster to.
        src_crs: str, optional
            The source crs to transform the raster from. If None, 
            get the projection from src_raster.

        Returns
        -------
        dst_raster: Raster
            Reprojected result.

        Examples
        -------- 
        >>> import TronGisPy as tgp
        >>> src_raster_fp = tgp.get_testing_fp()
        >>> src_raster = tgp.read_raster(src_raster_fp)
        >>> print("project(before)", src_raster.projection)
        >>> dst_raster = src_raster.reproject(dst_crs='EPSG:4326', src_crs=None)
        >>> print("project(after)", tgp.wkt_to_epsg(dst_raster.projection))
        """
        src_ds = self.to_gdal_ds()
        if src_crs:
            dst_ds = gdal.Warp('', src_ds, srcSRS=src_crs, dstSRS=dst_crs, format='MEM')
        else:
            dst_ds = gdal.Warp('', src_ds, dstSRS=dst_crs, format='MEM')
            if dst_ds is None:
                try:
                    src_crs = "EPSG:" + str(tgp.wkt_to_epsg(self.projection))
                    dst_ds = gdal.Warp('', src_ds, srcSRS=src_crs, dstSRS=dst_crs, format='MEM')
                    assert dst_ds is not None, "Please provide src_crs since the raster does not contain valid crs information."
                    print("Please provide src_crs to accelerate the process.")
                except:
                    assert False, "Please provide src_crs since the raster does not contain crs information."

        dst_raster = tgp.read_gdal_ds(dst_ds)
        return dst_raster

    def remap_by_ref_raster(self, ref_raster):
        """remap the raster on to another canvas drew by referenced raster.

        Parameters
        ----------
        ref_raster: Raster
            The referenced raster. The extent, projection and resolution will be used.

        Returns
        -------
        dst_raster: Raster
            Remapped result.

        """
        src_ds = self.to_gdal_ds()
        src_projection, src_gdaldtype = self.projection, self.gdaldtype
        ref_geo_transform, ref_projection = ref_raster.geo_transform, ref_raster.projection
        output_bounds = ref_raster.extent_for_gdal # minX, minY, maxX, maxY

        x_res, y_res = ref_geo_transform[1], ref_geo_transform[5]
        output_type = src_gdaldtype
        dst_ds = gdal.Warp('', src_ds, xRes=x_res, yRes=y_res, outputBounds=output_bounds, format='MEM', srcSRS=src_projection, dstSRS=ref_projection, outputType=output_type, resampleAlg='bilinear')
        dst_raster = tgp.read_gdal_ds(dst_ds)
        return dst_raster

    def hist(self, norm=False, clip_percentage=None, log=False, bands=None, ax=None, title=None, figsize=None):
        """plot digital value histgram (distribution) of raster object.

        Parameters
        ----------
        norm: bool, optional, default: False
            Normalize the image for showing.
        clip_percentage: tuple of float, optional
            The percentage to cut the data in head and tail e.g. (0.02, 0.98)
        log: bool, optional, default: False
            Use the log value to plot the histgram.
        bands: list, optional
            Which bands to plot. if bands==None, use values from all bands.
        ax: matplotlib.axes._subplots.AxesSubplot, optional
            On which ax the raster will be plot.
        title: str, optional
            The title of the histgram.
        figsize: tuple, optional
            Width and height of the histgram e.g.(10, 10).
        """
        if bands is None:
            data = self.data[self.data != self.no_data_value].flatten()
        else:
            data = self.data[:, :, bands]
            data = data[data != self.no_data_value].flatten()

        # clip_percentage
        if clip_percentage is not None:
            assert len(clip_percentage) == 2, "clip_percentage two element tuple"
            idx_st = int(len(data.flatten()) * clip_percentage[0])
            idx_end = int(len(data.flatten()) * clip_percentage[1])
            X_sorted = np.sort(data.flatten())
            data_min = X_sorted[idx_st]
            data_max = X_sorted[idx_end]
            data[data<data_min] = data_min
            data[data>data_max] = data_max
            
        # log
        if log:
            data = np.log(data)

        # normalize
        if norm:
            data = tgp.Normalizer().fit_transform(data, clip_percentage=clip_percentage)

        if ax is not None:
            ax.hist(data)
            ax.set_title(title)
        else:
            if figsize is not None:
                plt.figure(figsize=figsize)
            plt.hist(data)
            plt.title(title)
            plt.show()

    @property
    def cache_data_for_plot(self):
        """cache the processed data (norm & fill_na) for plotting

        Returns
        -------
        cache_data: `numpy.array`. Processed data.
        """
        return self.__cache_data_for_plot

    @cache_data_for_plot.setter
    def cache_data_for_plot(self, cache_data_for_plot):
        """set value for cache_data_for_plot"""
        if cache_data_for_plot is not None:
            self.__cache_data_for_plot = cache_data_for_plot
        else:
            self.__cache_data_for_plot = None

    def plot(self, flush_cache=True, norm=True, clip_percentage=(0.02, 0.98), clip_min_max=None, log=False, rescale_percentage=None, bands=None, ax=None, title=None, cmap=None, figsize=None):
        """Plot the raster object.

        Parameters
        ----------
        flush_cache: bool. 
            Flush the cached processed result for quick plotting.
        norm: bool, optional, default: True
            Normalize the image for showing.
        clip_percentage: tuple of float, optional, default: (0.02, 0.98)
            The percentage to cut the data in head and tail e.g. (0.02, 0.98)
        log: bool, optional, default: False
            Get the log value of data to show the image.
        rescale_percentage: float, optional
            The percentage to recale (resize) the image for efficient showing.
        bands: list, optional
            Which bands to plot. Length of bands should be 1, 3 or 4.
            If 3 bands is used, each of them will be defined as rgb bands. 
            If the forth band is used, it will be the opacity value.
        ax: matplotlib.axes._subplots.AxesSubplot, optional
            On which ax the raster will be plot.
        title: str, optional
            The title of the histgram.
        cmap: str, optional
            See Colormaps in Matplotlib https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html.
        figsize: tuple, optional
            Width and height of the histgram e.g.(10, 10).
        """
        if bands is None:
           bands = [0, 1, 2] if self.bands >= 3 else [0]
        
        assert type(bands) is list, "type of bands should be list"
        assert len(bands) in [1, 3, 4], "length of bands should be 1, 3 or 4"

        if (self.cache_data_for_plot is None) or flush_cache:
            # reshape to valid shape for matplotlib
            if len(bands) == 1:
                data = self.data[:, :, bands[0]]
            else:
                data = self.data[:, :, bands]

            # deal with no data
            data = data.astype(np.float)
            data[data == self.no_data_value] = np.nan

            # detect single value
            if len(np.unique(data[~np.isnan(data)])) == 1:
                norm = False
                log = False
                clip_percentage = None

            # clip_percentage & clip_min_max
            if clip_percentage is not None:
                assert not (clip_percentage is not None and clip_min_max is not None), "should not set clip_percentage and clip_min_max at the same time!"
                assert len(clip_percentage) == 2, "clip_percentage two element tuple"
                data = tgp.Normalizer().clip_by_percentage(data, clip_percentage=clip_percentage)
            elif clip_min_max is not None:
                assert len(clip_min_max) == 2, "clip_min_max two element tuple"
                data = tgp.Normalizer().clip_by_min_max(data, min_max=clip_min_max)

            # log
            if log:
                if data[~np.isnan(data)].min() < 0: 
                    data -= data[~np.isnan(data)].min()
                data[~np.isnan(data)] = np.log(data[~np.isnan(data)] + 10**-6)

            # normalize
            if norm:
                data = tgp.Normalizer().fit_transform(data)

            if rescale_percentage is not None:
                import cv2
                data = cv2.resize(data, (int(self.cols*rescale_percentage), int(self.rows*rescale_percentage)))

            self.cache_data_for_plot = data
        else:
            data = self.cache_data_for_plot

        # flip xy
        c, a, b, f, d, e = self.geo_transform
        # lng = a * col + b * row + c
        # lat = d * col + e * row + f
        # a = d(lng) / d(col)
        # b = d(lng) / d(row)
        # d = d(lat) / d(col)
        # e = d(lat) / d(row)
        if (np.abs(a) > np.abs(d)) and (np.abs(e) > np.abs(b)):
            lng_res, lat_res = a, e
        elif (np.abs(d) > np.abs(a)) and (np.abs(b) > np.abs(e)): # if d > a, 1 col move contribute more on lat, less on lng
            data = np.flipud(np.rot90(data))
            lng_res, lat_res = d, b
        else:
            assert False, "not acceptable geotransform"
        if lng_res < 0: # general lng_res > 0
            data = np.fliplr(data)
        if lat_res > 0: # general lat_res < 0
            data = np.flipud(data)

        # plotting
        if ax is not None:
            img = ax.imshow(data, extent=self.extent_for_plot, cmap=cmap)
            ax.set_title(title)
        else:
            if figsize is not None:
                plt.figure(figsize=figsize)
            img = plt.imshow(data, extent=self.extent_for_plot, cmap=cmap)
            plt.title(title)
            plt.show()
        return img
    


import gdal
import numpy as np
import TronGisPy as tgp

def dem_to_hillshade(src_raster, band=0, alg='Horn', azimuth=315, altitude=45):
    """Calculate the hillshade for the DEM.

    Parameters
    ----------
    src_raster : Raster
        The dem used to calculate the hillshade.
    band : int, optional, default: 0
        source band number to use.
    alg : {'ZevenbergenThorne' or 'Horn'}, optional, default: Horn
        The literature suggests Zevenbergen & Thorne to be more suited to smooth landscapes, 
        where Horn’s formula to perform better on rougher terrain.
    azimuth : int, optional, default 315
        Azimuth of the light, in degrees. 0 if it comes from the top of the raster, 90 from the east, ... 
        The default value, 315, should rarely be changed as it is the value generally used to generate shaded maps.
    altitude : int, optional, default 45
        Altitude of the light, in degrees. 90 if the light comes from above the DEM, 0 if it is raking light.

    Returns
    -------
    dst_raster: Raster
        Hillshade calculated from the DEM.
    """
    options = dict(band=band+1, alg=alg,  azimuth=azimuth, altitude=altitude, format='MEM')
    ds_src = src_raster.to_gdal_ds()
    ds = gdal.DEMProcessing('', ds_src, 'hillshade', **options)
    dst_raster = tgp.read_gdal_ds(ds)
    return dst_raster

def dem_to_slope(src_raster, band=0, alg='Horn', slope_format='degree'):
    """Calculate the slope for the DEM.

    Parameters
    ----------
    src_raster : Raster
        The dem used to calculate the slope.
    band : int, optional, default: 0
        source band number to use.
    alg : {'ZevenbergenThorne' or 'Horn'}, optional, default: Horn
        The literature suggests Zevenbergen & Thorne to be more suited to smooth landscapes, 
        where Horn’s formula to perform better on rougher terrain.
    slope_format: {"degree" or "percent"}, optional, default degree
        The format of the slope.

    Returns
    -------
    dst_raster: Raster
        Slope calculated from the DEM.
    """
    options = dict(band=band+1, alg=alg,  slopeFormat=slope_format, format='MEM')
    ds_src = src_raster.to_gdal_ds()
    ds = gdal.DEMProcessing('', ds_src, 'slope', **options)
    dst_raster = tgp.read_gdal_ds(ds)
    return dst_raster

def dem_to_aspect(src_raster, band=0, alg='Horn', trigonometric=False):
    """Calculate the aspect for the DEM.

    Parameters
    ----------
    src_raster : Raster
        The dem used to calculate the aspect.
    band : int, optional, default: 0
        source band number to use.
    alg : {'ZevenbergenThorne' or 'Horn'}, optional, default: Horn
        The literature suggests Zevenbergen & Thorne to be more suited to smooth landscapes, 
        where Horn’s formula to perform better on rougher terrain.
    trigonometric: bool, optional, default: False
        whether to return trigonometric angle instead of azimuth. 
        Thus 0deg means East, 90deg North, 180deg West, 270deg South.

    Returns
    -------
    dst_raster: Raster
        Aspect calculated from the DEM.
    """
    options = dict(band=band+1, alg=alg,  trigonometric=trigonometric, format='MEM')
    ds_src = src_raster.to_gdal_ds()
    ds = gdal.DEMProcessing('', ds_src, 'aspect', **options)
    dst_raster = tgp.read_gdal_ds(ds)
    return dst_raster

def dem_to_TRI(src_raster, band=0, alg='Horn'):
    """Calculate the terrain ruggedness index (TRI) for the DEM.

    Parameters
    ----------
    src_raster : Raster
        The dem used to calculate the TRI.
    band : int, optional, default: 0
        source band number to use.
    alg : {'ZevenbergenThorne' or 'Horn'}, optional, default: Horn
        The literature suggests Zevenbergen & Thorne to be more suited to smooth landscapes, 
        where Horn’s formula to perform better on rougher terrain.

    Returns
    -------
    dst_raster: Raster
        TRI calculated from the DEM.
    """
    options = dict(band=band+1, alg=alg, format='MEM')
    ds_src = src_raster.to_gdal_ds()
    ds = gdal.DEMProcessing('', ds_src, 'TRI', **options)
    dst_raster = tgp.read_gdal_ds(ds)
    return dst_raster

def dem_to_TPI(src_raster, band=0, alg='Horn'):
    """Calculate the topographic position index (TPI) for the DEM.

    Parameters
    ----------
    src_raster : Raster
        The dem used to calculate the TPI.
    band : int, optional, default: 0
        source band number to use.
    alg : {'ZevenbergenThorne' or 'Horn'}, optional, default: Horn
        The literature suggests Zevenbergen & Thorne to be more suited to smooth landscapes, 
        where Horn’s formula to perform better on rougher terrain.

    Returns
    -------
    dst_raster: Raster
        TPI calculated from the DEM.
    """
    options = dict(band=band+1, alg=alg, format='MEM')
    ds_src = src_raster.to_gdal_ds()
    ds = gdal.DEMProcessing('', ds_src, 'TPI', **options)
    dst_raster = tgp.read_gdal_ds(ds)
    return dst_raster

def dem_to_roughness(src_raster, band=0, alg='Horn'):
    """Calculate the roughness for the DEM.

    Parameters
    ----------
    src_raster : Raster
        The dem used to calculate the roughness.
    band : int, optional, default: 0
        source band number to use.
    alg : {'ZevenbergenThorne' or 'Horn'}, optional, default: Horn
        The literature suggests Zevenbergen & Thorne to be more suited to smooth landscapes, 
        where Horn’s formula to perform better on rougher terrain.

    Returns
    -------
    dst_raster: Raster
        roughness calculated from the DEM.
    """
    options = dict(band=band+1, alg=alg, format='MEM')
    ds_src = src_raster.to_gdal_ds()
    ds = gdal.DEMProcessing('', ds_src, 'Roughness', **options)
    dst_raster = tgp.read_gdal_ds(ds)
    return dst_raster

#TODO
# 1. Plan Curvature, Profile Curvature
# 2. LS-Factor
#==================================



import os
import gdal
import ogr, osr
import numpy as np
import TronGisPy as tgp
import geopandas as gpd
from collections import Counter
from TronGisPy import ShapeGrid
from shapely.geometry import Point, MultiPolygon, LineString, MultiLineString, Polygon

def clip_tif_by_shp(src_tif_path, src_shp_path, dst_tif_path):
    result = gdal.Warp(dst_tif_path,
                       src_tif_path,
                       cutlineDSName=src_shp_path,
                       cropToCutline=True)
    result = None

def clip_shp_by_shp(src_shp_path, clipper_shp_path, dst_shp_path):
    df_src = gpd.read_file(src_shp_path)
    df_clipper = gpd.read_file(clipper_shp_path)
    
    assert len(set([g.geom_type for g in  df_src['geometry']])) == 1, "geometry in the src_shp should have the same geom_type" 
    assert (len(set([g.geom_type for g in  df_clipper['geometry']])) == 1) and (df_clipper['geometry'].iloc[0].geom_type == 'Polygon'), "geom_type in the clipper_shp be Polygon" 

    geom_type = df_src['geometry'].iloc[0].geom_type
    if geom_type in ['Point', 'MultiPoint']:
        df_dst_shp = gpd.sjoin(df_src, df_clipper, how='inner')
    elif geom_type in ['Polygon', 'MultiPolygon']:
        df_dst_shp = gpd.overlay(df_src, df_clipper, how='intersection')
    elif geom_type in ['LineString', 'MultiLineString']:
        # TODO
        ## ogr solution: https://pcjericks.github.io/py-gdalogr-cookbook/vector_layers.html#get-geometry-from-each-feature-in-a-layer
        ## geopandas solution: MultiLineString intersection will in trounble
        # poly_to_be_clipped_path = get_testing_fp('poly_to_be_clipped')
        # point_to_be_clipped_path = get_testing_fp('point_to_be_clipped')
        # shp_clipper_path = get_testing_fp('shp_clipper')
        # df_poly = gpd.read_file(poly_to_be_clipped_path)
        # df_point = gpd.read_file(point_to_be_clipped_path)
        # lines = [np.stack([np.ones((4))*i, np.arange(1, 5)]).T for i in [1,3,7]] + \
        #         [np.stack([np.ones((2))*i, np.arange(3, 5)]).T for i in [0.5,2,6]]
        # df_line = gpd.GeoDataFrame(geometry=[LineString(line) for line in lines])
        # df_multiline = gpd.GeoDataFrame(geometry=[MultiLineString(lines)])
        # df_clipper = gpd.read_file(shp_clipper_path)
        # 
        # fig, ax = plt.subplots(1,1)
        # df_clipper.plot(ax=ax)
        # df_line.plot(ax=ax, color='red')
        # 
        # df_line.intersection(df_clipper) 
        ## 1. num of clipper polygons will have different return rows
        ## 2. not intersect line will return None or "GEOMETRYCOLLECTION EMPTY"
        # df_multiline.intersection(df_clipper) 
        ## some part of WultiLine willdisappear
        # assert False, "We temporarily does not support for line strings clipping"
        multi_lines = []
        for line in df_src['geometry']:
            lines = []
            for poly in df_clipper['geometry']:
                line_intersection = line.intersection(poly)
                if not line_intersection.is_empty:
                    if line_intersection.geom_type == 'MultiLineString':
                        lines.extend(line.intersection(poly))
                    elif line_intersection.geom_type == 'LineString':
                        lines.append(line.intersection(poly))
            multi_lines.append(MultiLineString(lines))
        df_dst_shp = df_src.copy()
        df_dst_shp['geometry'] = multi_lines
        df_dst_shp.dropna(inplace=True)
    else:
        assert False, "geom_type must be Point, MultiPoint, Polygon, MultiPolygon, LineString or MultiLineString!"

    df_dst_shp.to_file(dst_shp_path)

def tif_composition(ref_tif_path, src_tif_paths, dst_tif_path, dst_tif_dtype_gdal=None):
    """
    ref_tif_path: should be used to create the canvas with final coordinate system, geo_transform and projection, 
    src_tif_paths: should be in list type with elements with full path of tif images.
    dst_tif_path: output file path
    """
    # get geo info
    rows, cols, bands, geo_transform, projection, dtype_gdal, no_data_value, metadata = tgp.get_raster_info(ref_tif_path)
    if dst_tif_dtype_gdal:
        dtype_gdal = dst_tif_dtype_gdal
        
    # cal bands count
    bands_for_each_tif = [tgp.get_raster_info(tif_path)[2] for tif_path in src_tif_paths]
    bands = sum(bands_for_each_tif)

    # bands compositions: create new tif
    dst_ds = gdal.GetDriverByName('GTiff').Create(dst_tif_path, cols, rows, bands, dtype_gdal)
    dst_ds.SetGeoTransform(geo_transform)
    dst_ds.SetProjection(projection)
    
    # bands compositions: write bands
    band_num = 1
    for tif_path, bands_for_the_tif in zip(src_tif_paths, bands_for_each_tif):
        nparr = tgp.get_raster_data(tif_path)
        for band_num_for_the_tif in range(bands_for_the_tif):
            band = dst_ds.GetRasterBand(band_num)
            band.WriteArray(nparr[:, :, band_num_for_the_tif], 0, 0)
            band.FlushCache()
            if no_data_value:
                band.SetNoDataValue(no_data_value)
            band_num += 1
    dst_ds = None

def refine_resolution(src_tif_path, dst_tif_path, dst_resolution, resample_alg='near'):
    """
    near: nearest neighbour resampling (default, fastest algorithm, worst interpolation quality).
    bilinear: bilinear resampling.
    cubic: cubic resampling.
    cubicspline: cubic spline resampling.
    lanczos: Lanczos windowed sinc resampling.
    average: average resampling, computes the weighted average of all non-NODATA contributing pixels.
    mode: mode resampling, selects the value which appears most often of all the sampled points.
    """
    result = gdal.Warp(dst_tif_path, src_tif_path, xRes=dst_resolution, yRes=dst_resolution, resampleAlg=resample_alg)
    result = None

def rasterize_layer(src_shp_path, dst_tif_path, ref_tif_path, use_attribute=None, all_touched=False, gdaldtype=None, no_data_value=None):
    """
    src_shp_path: rasterize which shp.
    dst_tif_path: rasterize output, should be in tiff type.
    ref_tif_path: the geo information reference raster.
    use_attribute: use thich attribute of the shp as raster value.
    """
    # Open your shapefile
    gdf_shp = gpd.read_file(src_shp_path)
    if not use_attribute:
        use_attribute = 'positive'
        gdf_shp[use_attribute] = 1


    # Create the destination raster data source
    # pixelWidth = pixelHeight = 2 # depending how fine you want your raster
    # x_min, x_max, y_min, y_max = source_layer.GetExtent()
    # cols = int((x_max - x_min) / pixelHeight)
    # rows = int((y_max - y_min) / pixelWidth)
    # geoTransform = (x_min, pixelWidth, 0, y_min, 0, pixelHeight)
    ref_tif_ds = gdal.Open(ref_tif_path)
    ref_tif_cols, ref_tif_rows = ref_tif_ds.RasterXSize, ref_tif_ds.RasterYSize
    ref_tif_geo_transofrm = ref_tif_ds.GetGeoTransform()

    rasterize_raster = ShapeGrid.rasterize_layer(gdf_shp, ref_tif_rows, ref_tif_cols, ref_tif_geo_transofrm, use_attribute, all_touched, no_data_value=no_data_value)
    rasterize_raster.to_file(dst_tif_path)


def vectorize_layer(src_tif_path, dst_shp_path, field_name='value', band_num=0, multipolygon=False):
    """band_num start from 1"""
    src_ds = gdal.Open(src_tif_path)
    srcband = src_ds.GetRasterBand(band_num+1)
    src_srs=osr.SpatialReference(wkt=src_ds.GetProjection())        
    drv = ogr.GetDriverByName("ESRI Shapefile")
    dst_ds = drv.CreateDataSource(dst_shp_path)
    layer_name = os.path.split(dst_shp_path)[-1].replace(".shp", "")
    dst_layer = dst_ds.CreateLayer(layer_name, srs=src_srs)
    dst_layer.CreateField(ogr.FieldDefn(field_name, ogr.OFTInteger))
    dst_field = dst_layer.GetLayerDefn().GetFieldIndex(field_name)
    gdal.Polygonize(srcband, None, dst_layer, dst_field, [], callback=None)

    src_ds = None
    dst_ds = None

    if multipolygon:
        df_shp = gpd.read_file(dst_shp_path)
        multi_polygons = df_shp.groupby(field_name)['geometry'].apply(list).apply(MultiPolygon)
        values = df_shp.groupby(field_name)[field_name].first()
        df_shp = gpd.GeoDataFrame(geometry=multi_polygons)
        df_shp[field_name] = values
        df_shp.to_file(dst_shp_path, index=False)

def raster_pixel_to_polygon(src_tif_path, dst_shp_path, all_bands_as_feature=False, crs=None, return_gdf=False):
    """
    crs should be dict type 'epsg:<epsg_code>', e.g. 'epsg:4326'
    """
    rows, cols, bands, geo_transform, projection, dtype_gdal, no_data_value, metadata = tgp.get_raster_info(src_tif_path)
    X = tgp.get_raster_data(src_tif_path)
    idxs = np.where(np.ones_like(X[:,:,0], dtype=np.bool))
    rows = []

    for row_idx, col_idx in zip(*idxs):
        row = {}
        npidx = (row_idx, col_idx)
        row['geometry'] = Polygon(tgp.npidxs_to_coord_polygons([npidx], geo_transform)[0])
        if all_bands_as_feature:
            for i in range(X.shape[2]):
                row['band'+str(i+1)] = X[row_idx, col_idx, i]
        rows.append(row)
    df_shp = gpd.GeoDataFrame(rows, geometry='geometry')
    if crs:
        df_shp.crs = crs
    if return_gdf:
       return df_shp
    else:
        df_shp.to_file(dst_shp_path)


def reproject(src_tif_path, dst_tif_path, dst_crs='EPSG:4326', src_crs=None):
    if src_crs:
        gdal.Warp(dst_tif_path, src_tif_path, srcSRS=src_crs, dstSRS=dst_crs)
    else:
        gdal.Warp(dst_tif_path, src_tif_path, dstSRS=dst_crs)


def remap_tif(src_tif_path, dst_tif_path, ref_tif_path):
    src_projection, src_gdaldtype = tgp.get_raster_info(src_tif_path, ['projection', 'gdaldtype'])
    ref_geo_transform, ref_projection = tgp.get_raster_info(ref_tif_path, ['geo_transform', 'projection'])
    output_bounds = tgp.get_raster_extent(ref_tif_path, 'gdal') # (minX, minY, maxX, maxY)

    x_res, y_res = ref_geo_transform[1], ref_geo_transform[5]
    output_type = src_gdaldtype
    gdal.Warp(dst_tif_path, src_tif_path, outputBounds=output_bounds,
                                            xRes=x_res,
                                            yRes=y_res,
                                            outputType=output_type,
                                            srcSRS=src_projection,
                                            dstSRS=ref_projection)



#TODO
# 1. raster pixel to points



import numpy as np

def _get_M(omega, phi, kappa):
    """Collinearity Equation"""
    cos, sin = np.cos(omega), np.sin(omega)
    Mo = np.array([[    1,    0,    0], 
                   [    0,  cos,  sin],
                   [    0, -sin,  cos]])
    cos, sin = np.cos(phi), np.sin(phi)
    Mp = np.array([[  cos,    0, -sin],
                   [    0,    1,    0],
                   [  sin,    0,  cos]])
    cos, sin = np.cos(kappa), np.sin(kappa)
    Mk = np.array([[  cos,  sin,    0],
                   [ -sin,  cos,    0],
                   [    0,    0,    1]])
    M = np.matmul(np.matmul(Mk, Mp), Mo)
    return M

def _convert_npidxs_to_imxys(npidxs, rows=13824, cols=7680, pixel_size=0.012):
    """
    image space linear transform
    pixel_size = sensor_width/image_cols = 92.16/7680 = sensor_height/image_rows = 165.8885/13824  # unit: mm/px
    npidxs
    """
    col_zoomout, row_zoomout = pixel_size, -pixel_size
    row_idxs, col_idxs = npidxs.T
    xs = (col_idxs - (cols/2)) * col_zoomout
    ys = (row_idxs - (rows/2)) * row_zoomout
    imxys = np.array([xs, ys]).T
    return imxys

def _convert_imxyzs_to_npidxs(imxyzs, rows=13824, cols=7680, pixel_size=0.012):
    """
    image space linear transform
    pixel_size = sensor_width/image_cols = 92.16/7680 = sensor_height/image_rows = 165.8885/13824  # unit: mm/px
    """
    col_zoomin, row_zoomin = 1/pixel_size, -1/pixel_size
    xs, ys = imxyzs.T[0], imxyzs.T[1]
    col_idxs = col_zoomin * xs + (cols/2)
    row_idxs = row_zoomin * ys + (rows/2)
    npidxs = np.array([row_idxs, col_idxs]).T
    return npidxs

def project_npidxs_to_XYZs(P_npidxs, P_Z, aerotri_params, return_k=False):
    """
    idxs conversion from image space and object space.
    P_xyzs: xyzs of points in image space.
    P_Z: assume height.
    return_k: assume height.
    """
    # aerotri_params
    opk, L_XYZ, rows, cols, focal_length, pixel_size = aerotri_params
    # L_XYZ = np.expand_dims(L_XYZ, axis=-1)

    # calculate P_imxyzs from P_npidxs
    P_xys = _convert_npidxs_to_imxys(np.array(P_npidxs), rows, cols, pixel_size) # (n points, 2 dim)
    P_xyzs = np.concatenate([P_xys, np.full((len(P_xys), 1), -focal_length)], axis=1).T # add -focal_length as new dim to (3 dim, n points)

    # calculate P_XYZs from P_imxyzs
    M, L_Z = _get_M(*opk), L_XYZ[2]
    mt_xyzs = np.matmul(M.T, P_xyzs)
    ks = mt_xyzs[2] / (P_Z - L_Z) # convert from (L_Z - P_Z) = P_xyzs[2] / k, and query for k
    P_XYZs = ((mt_xyzs / ks) + np.expand_dims(L_XYZ, axis=-1)).T # 4 points, 3 dim
    if not return_k:
        return P_XYZs
    else:
        return P_XYZs, ks


def project_XYZs_to_npidxs(P_XYZs, aerotri_params, return_k=False):
    """
    idxs conversion from object space and image space.
    P_XYZs: XYZs of points in object space.
    aerotri_params: opk, L_XYZ, rows, cols, focal_length, pixel_size
        opk: omega, phi, kappa.
        L_XYZ: the location of the camera.
        focal_length: dmc image is 120 (m).
        pixel_size: dmc image is sensor_width/image_cols = sensor_height/image_rows = 0.012 (mm).
    """
    # aerotri_params
    opk, L_XYZ, rows, cols, focal_length, pixel_size = aerotri_params
    
    # calculate P_imxyzs from P_XYZs
    M = _get_M(*opk)
    P_XYZs = np.array(P_XYZs).T  # =>P_XYZs.shape==(3, None)
    L_XYZ = np.array(L_XYZ).reshape(3, 1) # =>L_XYZ.shape==(3, 1)
    dXYZs = (P_XYZs-L_XYZ) # =>xyz.shape==(3, None)
    m_XYZs = np.matmul(M, dXYZs)
    ks = -focal_length / m_XYZs[2]
    P_xyzs = (ks * m_XYZs).T # (n points, 3 dim)
    
    # calculate P_npidxs from P_xyzs
    P_npidxs = _convert_imxyzs_to_npidxs(P_xyzs, rows, cols, pixel_size)
    if not return_k:
        return P_npidxs
    else:
        return P_npidxs, ks



def get_img_pixels_XYZs(npidxs, aerotri_params, k=-1):
    """
    idxs conversion from image space and object space.
    P_xyzs: xyzs of points in image space.
    """
    # aerotri_params
    opk, L_XYZ, rows, cols, focal_length, pixel_size = aerotri_params

    # calculate P_imxyzs from npidxs
    P_xys = _convert_npidxs_to_imxys(np.array(npidxs), rows, cols, pixel_size) # (n points, 2 dim)
    P_xyzs = np.concatenate([P_xys, np.full((len(P_xys), 1), -focal_length)], axis=1).T # add -focal_length as new dim to (3 dim, n points)

    # calculate P_XYZs from P_imxyzs
    M, L_Z = _get_M(*opk), L_XYZ[2]
    mt_xyzs = np.matmul(M.T, P_xyzs)
    P_XYZs = (L_XYZ - (mt_xyzs / k).T) # 4 points, 3 dim
    return P_XYZs


import os
import cv2
import ogr
import gdal
import pyproj
import shapely
import numpy as np
import pandas as pd
import geopandas as gpd
import TronGisPy as tgp
from TronGisPy import CRS
def get_rasterize_layer_params(src_vector, res=5):
    """Get params for rasterize_layer if you don't have a grid system. 

    Parameters
    ----------
    src_vector: Geopandas.GeoDataFrame
        The vector data to be rasterize.
    res: resolution
        The resolution (in meters) of the grid.

    Returns
    -------
    rows: int
        Target rasterized image's rows.
    cols: int
        Target rasterized image's cols.
    geo_transform: tuple
        Target rasterized image's geo_transform which is the affine parameter.
    """
    xmin, ymin, xmax, ymax = src_vector.total_bounds 
    geo_transform = (xmin, res, 0, ymax, 0, -res)
    cols = int((xmax - xmin) / res)
    rows = int((ymax - ymin) / res)
    return rows, cols, geo_transform

def rasterize_layer(src_vector, rows, cols, geo_transform, use_attribute, all_touched=False, no_data_value=0):
    """Rasterize vector data. Get the cell value in defined grid (rows, cols, geo_transform)
    from its overlapped polygon.

    Parameters
    ----------
    src_vector: Geopandas.GeoDataFrame
        Which vector data to be rasterize.
    rows: int
        Target rasterized image's rows.
    cols: int
        Target rasterized image's cols.
    geo_transform: tuple
        Target rasterized image's geo_transform which is the affine parameter.
    use_attribute: str
        The column to use as rasterized image value.
    all_touched: bool, optioonal, default: False
        Pixels that touch (not overlap over 50%) the polygon will be assign the use_attribute value of the polygon.
    no_data_value: int or float
        The pixels not covered by any polygon will be filled no_data_value.

    Returns
    -------
    raster: Raster. 
        Rasterized result.

    Examples
    -------- 
    >>> import geopandas as gpd
    >>> import TronGisPy as tgp 
    >>> from TronGisPy import ShapeGrid
    >>> from matplotlib import pyplot as plt
    >>> ref_raster_fp = tgp.get_testing_fp('satellite_tif') # get the geoinfo from the raster
    >>> src_vector_fp = tgp.get_testing_fp('satellite_tif_clipper') # read source shapefile as GeoDataFrame
    >>> src_vector = gpd.read_file(src_vector_fp)
    >>> src_vector['FEATURE'] = 1 # make the value to fill in the raster cell
    >>> rows, cols, geo_transform = tgp.get_raster_info(ref_raster_fp, ['rows', 'cols', 'geo_transform'])
    >>> raster = ShapeGrid.rasterize_layer(src_vector, rows, cols, geo_transform, use_attribute='FEATURE', no_data_value=99)
    >>> fig, (ax1, ax2) = plt.subplots(1,2) # plot the result
    >>> tgp.read_raster(ref_raster_fp).plot(ax=ax1)
    >>> src_vector.plot(ax=ax1)
    >>> ax1.set_title('polygon with ref_raster')
    >>> raster.plot(ax=ax2)
    >>> ax2.set_title('rasterized image')
    >>> plt.show()
    """
    # Open your shapefile
    assert type(src_vector) is gpd.GeoDataFrame, "src_vector should be GeoDataFrame type."
    assert use_attribute in src_vector.columns, "attribute not exists in src_vector."
    gdaldtype = tgp.npdtype_to_gdaldtype(src_vector[use_attribute].dtype)
    # projection = src_vector.crs.to_wkt() if src_vector.crs is not None else None
    projection = pyproj.CRS(src_vector.crs).to_wkt() if src_vector.crs is not None else None
    src_shp_ds = ogr.Open(src_vector.to_json())
    src_shp_layer = src_shp_ds.GetLayer()

    # Create the destination raster data source
    ds = tgp.write_gdal_ds(bands=1, cols=cols, rows=rows, geo_transform=geo_transform, gdaldtype=gdaldtype, no_data_value=no_data_value)

    # set it to the attribute that contains the relevant unique
    options = ["ATTRIBUTE="+use_attribute]
    if all_touched:
        options.append('ALL_TOUCHED=TRUE')
    gdal.RasterizeLayer(ds, [1], src_shp_layer, options=options) # target_ds, band_list, source_layer, options = options

    data = ds.GetRasterBand(1).ReadAsArray()
    raster = tgp.Raster(data, geo_transform, projection, gdaldtype, no_data_value)
    return raster

def rasterize_layer_by_ref_raster(src_vector, ref_raster, use_attribute, all_touched=False, no_data_value=0):
    """Rasterize vector data. Get the cell value in defined grid of ref_raster
    from its overlapped polygon.

    Parameters
    ----------
    src_vector: Geopandas.GeoDataFrame
        Which vector data to be rasterize.
    ref_raster: Raster
        Target rasterized image's rows, cols, and geo_transform.
    use_attribute: str
        The column to use as rasterized image value.
    all_touched: bool, optioonal, default: False
        Pixels that touch (not overlap over 50%) the polygon will be assign the use_attribute value of the polygon.
    no_data_value: int or float
        The pixels not covered by any polygon will be filled no_data_value.

    Returns
    -------
    raster: Raster. 
        Rasterized result.

    Examples
    -------- 
    >>> import geopandas as gpd
    >>> import TronGisPy as tgp 
    >>> from TronGisPy import ShapeGrid
    >>> from matplotlib import pyplot as plt
    >>> ref_raster_fp = tgp.get_testing_fp('satellite_tif') # get the geoinfo from the raster
    >>> src_vector_fp = tgp.get_testing_fp('satellite_tif_clipper') # read source shapefile as GeoDataFrame
    >>> src_vector = gpd.read_file(src_vector_fp)
    >>> src_vector['FEATURE'] = 1 # make the value to fill in the raster cell
    >>> ref_raster = tgp.read_raster(ref_raster_fp)
    >>> raster = ShapeGrid.rasterize_layer_by_ref_raster(src_vector, ref_raster, use_attribute='FEATURE', no_data_value=99)
    >>> fig, (ax1, ax2) = plt.subplots(1,2) # plot the result
    >>> tgp.read_raster(ref_raster_fp).plot(ax=ax1)
    >>> src_vector.plot(ax=ax1)
    >>> ax1.set_title('polygon with ref_raster')
    >>> raster.plot(ax=ax2)
    >>> ax2.set_title('rasterized image')
    >>> plt.show()
    """
    # Open your shapefile
    assert type(src_vector) is gpd.GeoDataFrame, "src_vector should be GeoDataFrame type."
    assert use_attribute in src_vector.columns, "attribute not exists in src_vector."
    rows, cols, geo_transform = ref_raster.rows, ref_raster.cols, ref_raster.geo_transform
    raster = rasterize_layer(src_vector, rows, cols, geo_transform, use_attribute=use_attribute, all_touched=all_touched, no_data_value=no_data_value)
    return raster

def vectorize_layer(src_raster, band_num=1, field_name='value', multipolygon=False):
    """Vectorize raster data to achieve an acceptable raster-to-vector conversion.

    Parameters
    ----------
    src_raster: Raster
        Which raster data to be vectorize.
    band_num: int
        Which band to be vectorized.
    field_name: str, optional, default: value
        Field to be generated in output vector data.
    multipolygon: bool, optional, default: False
        Combine the polygon with the same value to be a `shapely.geometry.MultiPolygon`.

    Returns
    -------
    vector: Geopandas.GeoDataFrame
        Vectorized result.

    Examples
    -------- 
    >>> import TronGisPy as tgp
    >>> from TronGisPy import ShapeGrid
    >>> from matplotlib import pyplot as plt
    >>> src_raster_fp = tgp.get_testing_fp('rasterized_image_1')
    >>> src_raster = tgp.read_raster(src_raster_fp)
    >>> df_shp = ShapeGrid.vectorize_layer(src_raster)
    >>> fig, ax = plt.subplots(1, 1) # plot the result
    >>> src_raster.plot(ax=ax)
    >>> df_shp.boundary.plot(ax=ax, color='red', linewidth=5)
    >>> plt.show()
    """
    src_ds = src_raster.to_gdal_ds()
    src_band = src_ds.GetRasterBand(band_num)

    drv = ogr.GetDriverByName("MEMORY")
    dst_ds = drv.CreateDataSource('memData')
    dst_layer = dst_ds.CreateLayer('vectorize_layer')
    dst_layer.CreateField(ogr.FieldDefn(field_name, ogr.OFTInteger)) # TODO: should not force convert ot integer
    dst_field = dst_layer.GetLayerDefn().GetFieldIndex(field_name)
    gdal.Polygonize(src_band, None, dst_layer, dst_field, [], callback=None)

    df_vectorized_rows = []
    for feature in dst_layer:
        field_value = feature.GetField(field_name)
        geometry = shapely.wkt.loads(feature.GetGeometryRef().ExportToWkt())
        df_vectorized_rows.append([field_value, geometry])
    df_vectorized = gpd.GeoDataFrame(df_vectorized_rows, columns=[field_name, 'geometry'], geometry='geometry', crs=src_raster.projection)
    src_ds = None
    dst_ds = None

    if multipolygon:
        from shapely.geometry import MultiPolygon
        multi_polygons = df_vectorized.groupby(field_name)['geometry'].apply(list).apply(MultiPolygon)
        values = df_vectorized.groupby(field_name)[field_name].first()
        df_vectorized = gpd.GeoDataFrame(geometry=multi_polygons)
        df_vectorized[field_name] = values

    return df_vectorized

# def clip_raster_with_polygon(src_raster, src_poly):
#     """Clip raster with polygon.

#     Parameters
#     ----------
#     src_raster: Raster
#         Which raster data to be clipped.
#     src_poly: Geopandas.GeoDataFrame
#         The clipper(clipping boundary).

#     Returns
#     -------
#     dst_raster: Raster 
#         Clipped result.

#     Examples
#     -------- 
#     >>> import geopandas as gpd
#     >>> import TronGisPy as tgp
#     >>> from TronGisPy import ShapeGrid
#     >>> from matplotlib import pyplot as plt
#     >>> src_raster_fp = tgp.get_testing_fp('satellite_tif')
#     >>> src_poly_fp = tgp.get_testing_fp('satellite_tif_clipper')
#     >>> src_raster = tgp.read_raster(src_raster_fp)
#     >>> src_poly = gpd.read_file(src_poly_fp)
#     >>> dst_raster = ShapeGrid.clip_raster_with_polygon(src_raster, src_poly)
#     >>> fig, (ax1, ax2) = plt.subplots(1, 2) # plot the result
#     >>> src_raster.plot(ax=ax1)
#     >>> src_poly.boundary.plot(ax=ax1)
#     >>> ax1.set_title('original image and clipper')
#     >>> dst_raster.plot(ax=ax2)
#     >>> ax2.set_title('clipped image')
#     >>> plt.show()
#     """
#     assert src_raster.geo_transform is not None, "src_raster.geo_transform should not be None"
#     src_ds = src_raster.to_gdal_ds()
#     temp_dir = tgp.create_temp_dir_when_not_exists()
#     src_shp_fp = os.path.join(temp_dir, 'src_poly.shp')
#     src_poly.to_file(src_shp_fp)
#     dst_ds = gdal.Warp('', src_ds, format= 'MEM', cutlineDSName=src_shp_fp, cropToCutline=True)
#     dst_raster = tgp.read_gdal_ds(dst_ds)
#     return dst_raster

def clip_raster_with_polygon(src_raster, src_poly, all_touched=False, no_data_value=0):
    """Clip raster with polygon.

    Parameters
    ----------
    src_raster: Raster
        Which raster data to be clipped.
    src_poly: Geopandas.GeoDataFrame
        The clipper(clipping boundary).

    Returns
    -------
    dst_raster: Raster 
        Clipped result.

    Examples
    -------- 
    >>> import geopandas as gpd
    >>> import TronGisPy as tgp
    >>> from TronGisPy import ShapeGrid
    >>> from matplotlib import pyplot as plt
    >>> src_raster_fp = tgp.get_testing_fp('satellite_tif')
    >>> src_poly_fp = tgp.get_testing_fp('satellite_tif_clipper')
    >>> src_raster = tgp.read_raster(src_raster_fp)
    >>> src_poly = gpd.read_file(src_poly_fp)
    >>> dst_raster = ShapeGrid.clip_raster_with_polygon(src_raster, src_poly)
    >>> fig, (ax1, ax2) = plt.subplots(1, 2) # plot the result
    >>> src_raster.plot(ax=ax1)
    >>> src_poly.boundary.plot(ax=ax1)
    >>> ax1.set_title('original image and clipper')
    >>> dst_raster.plot(ax=ax2)
    >>> ax2.set_title('clipped image')
    >>> plt.show()
    """
    assert src_raster.geo_transform is not None, "src_raster.geo_transform should not be None"
    src_poly_copy = src_poly.copy()
    src_poly_copy['value'] = 1
    src_poly_raster = rasterize_layer_by_ref_raster(src_poly_copy, src_raster, use_attribute='value', all_touched=all_touched, no_data_value=0)
    dst_raster = src_raster.copy()
    dst_raster.data[~(src_poly_raster.data[:, :, 0].astype(bool))] = no_data_value
    
    row_idxs, col_idxs, bands_idxs = np.where(src_poly_raster.data!=0)
    rmin, rmax, cmin, cmax = np.min(row_idxs), np.max(row_idxs), np.min(col_idxs), np.max(col_idxs)
    dst_raster.data = dst_raster.data[rmin:rmax+1, cmin:cmax+1]

    coords = tgp.npidxs_to_coords([(rmin, cmin)], src_raster.geo_transform)[0]
    geo_transform = np.array(dst_raster.geo_transform)
    geo_transform[[0, 3]] = coords
    dst_raster.geo_transform = geo_transform

    # src_ds = src_raster.to_gdal_ds()
    # temp_dir = tgp.create_temp_dir_when_not_exists()
    # src_shp_fp = os.path.join(temp_dir, 'src_poly.shp')
    # src_poly.to_file(src_shp_fp)
    # dst_ds = gdal.Warp('', src_ds, format= 'MEM', cutlineDSName=src_shp_fp, cropToCutline=True)
    # dst_raster = tgp.read_gdal_ds(dst_ds)
    return dst_raster

# split partitions
def __split_idxs_partitions(idxs, partitions, seed=None):
    if seed:
        np.random.seed(seed)
    partition_len = int(len(idxs) // partitions)
    parts = []
    for i in range(partitions - 1):
        part = np.random.choice(idxs, size=partition_len, replace=False)
        parts.append(part)
        idxs = list(set(idxs) - set(part))
    parts.append(idxs)
    return parts 

def clip_raster_with_multiple_polygons(src_raster, src_poly, partitions=10, return_raster=False, no_data_value=None, seed=None):
    """Clip raster with multiple polygons in the same shp as independent image.

    Parameters
    ----------
    src_raster: Raster
        Which raster data to be clipped.
    src_poly: Geopandas.GeoDataFrame
        The clipper(clipping boundary).
    partitions: int, default: 10
        The number of partitions used to split all polygons in diferent parts 
        and rasterize them in different iterations in order to avoid overlapping 
        when rasterizing.
    return_raster:bool, optional, default: False
        Return np.array if return_raster == False, else return tgp.Raster.
    no_data_value:int, optional
        Set no_data_value for clipped image. If None, use src_raster.no_data_value 
        as default. If src_raster.no_data_value is None, use 0 as default.
    seed: int, optional
        Seed to split partitions.

    Returns
    -------
    dst_raster: Raster 
        Clipped result.

    Examples
    -------- 
    >>> import numpy as np
    >>> import geopandas as gpd
    >>> import TronGisPy as tgp
    >>> from TronGisPy import ShapeGrid
    >>> from matplotlib import pyplot as plt
    >>> src_raster_fp = tgp.get_testing_fp('multiple_poly_clip_ras')
    >>> src_poly_fp = tgp.get_testing_fp('multiple_poly_clipper')
    >>> src_raster = tgp.read_raster(src_raster_fp)
    >>> src_shp = gpd.read_file(src_poly_fp)
    >>> clipped_imgs = ShapeGrid.clip_raster_with_multiple_polygons(src_raster, src_shp, return_raster=True)
    >>> fig, axes = plt.subplots(2, 5, figsize=(9, 6))
    >>> axes = axes.flatten()
    >>> for idx, ax in zip(np.arange(100, 100+10, 1), axes):
    >>>     clipped_imgs[idx].plot(ax=ax)
    >>> fig.suptitle("TestShapeGrid" + ": " + "test_clip_raster_with_multiple_polygons")
    >>> plt.show()
    """
    # init resource
    assert (len(src_poly) // partitions) < np.iinfo(np.int32).max, "Please increase partitions in order the gdal type overflow issue."
    df_poly_for_rasterize = src_poly.copy()
    partitions = len(src_poly) if len(src_poly) < partitions else partitions   
    df_poly_for_rasterize.loc[:, 'id'] = range(len(df_poly_for_rasterize))
    parts = __split_idxs_partitions(df_poly_for_rasterize['id'].values, partitions=partitions, seed=seed)
    if no_data_value is None:
        no_data_value = 0 if src_raster.no_data_value is None else src_raster.no_data_value

    # rasterize by its id and clipping
    clipped_imgs = []
    for ps_idx, ps in enumerate(parts): # deal with one part of poly in shp per loop: 1. rasterize => 2. find each poly in the shp
        # 1. rasterize: rasterize only df_plot['id'].isin(ps) (only id in the splitted shp)
        df_poly_for_rasterize_ps = pd.concat([df_poly_for_rasterize[df_poly_for_rasterize['id'] == p].copy() for p in ps])
        df_poly_for_rasterize_ps.loc[:, 'id_ps'] = np.array(range(len(df_poly_for_rasterize_ps)), dtype=np.int32)
        raster_poly_part = rasterize_layer(df_poly_for_rasterize_ps, src_raster.rows, src_raster.cols, src_raster.geo_transform, use_attribute='id_ps', all_touched=True, no_data_value=-1)
        
        for id_p in range(len(df_poly_for_rasterize_ps)):
            # 2. find each the location (in the raster) of each poly in the shp 
            coords = df_poly_for_rasterize_ps[df_poly_for_rasterize_ps['id_ps'] == id_p].total_bounds.reshape(2,2)
            npidxs = CRS.coords_to_npidxs(coords, src_raster.geo_transform)
            row_idxs_st, row_idxs_end, col_idxs_st, col_idxs_end = np.min(npidxs[:, 0]), np.max(npidxs[:, 0])+1, np.min(npidxs[:, 1]), np.max(npidxs[:, 1])+1
            clipped_img = src_raster.data[row_idxs_st:row_idxs_end, col_idxs_st:col_idxs_end].copy()
            ploy_mask = raster_poly_part.data[row_idxs_st:row_idxs_end, col_idxs_st:col_idxs_end, 0] == id_p
            if np.sum(ploy_mask) > 0:
                # generate clipped image
                clipped_img[~ploy_mask] = no_data_value
                if return_raster:
                    gt = np.array(src_raster.geo_transform)
                    gt[[0, 3]] = CRS.npidxs_to_coords([(row_idxs_st, col_idxs_st)], src_raster.geo_transform)[0]
                    clipped_img = tgp.Raster(clipped_img, tuple(gt), src_raster.projection, src_raster.gdaldtype, no_data_value, src_raster.metadata)
                clipped_imgs.append(clipped_img)
            else:
                clipped_imgs.append(None)
        
        # na_percentage = np.sum([c is None for c in clipped_imgs[-len(df_poly_for_rasterize_ps):]]) / len(df_poly_for_rasterize_ps)
        # if na_percentage != 0 : 
        #     print(ps_idx, na_percentage)
                
    clipped_imgs = [clipped_imgs[i] for i in np.argsort(np.hstack(parts))]
    return clipped_imgs

def clip_raster_with_extent(src_raster, extent):
    """Clip raster with extent.

    Parameters
    ----------
    src_raster: Raster
        Which raster data to be clipped.
    extent: tuple
        extent to clip the data with (xmin, ymin, xmax, ymax) format.

    Returns
    -------
    dst_raster: Raster. 
        Clipped result.

    Examples
    -------- 
    >>> import geopandas as gpd
    >>> import TronGisPy as tgp
    >>> from TronGisPy import ShapeGrid
    >>> from matplotlib import pyplot as plt
    >>> from shapely.geometry import Polygon
    >>> src_raster_fp = tgp.get_testing_fp('satellite_tif')
    >>> src_raster = tgp.read_raster(src_raster_fp)
    >>> ext = xmin, ymin, xmax, ymax = [329454.39272725, 2746809.43272727, 331715.57090906, 2748190.90181818]
    >>> dst_raster = ShapeGrid.clip_raster_with_extent(src_raster, ext)
    >>> fig, (ax1, ax2) = plt.subplots(1, 2) # plot the result
    >>> src_raster.plot(ax=ax1) 
    >>> ext_poly = [(xmin, ymin), (xmin, ymax), (xmax, ymax), (xmax, ymin)]
    >>> df_ext_poly = gpd.GeoDataFrame([Polygon(ext_poly)], columns=['geom'], geometry='geom')
    >>> df_ext_poly.boundary.plot(ax=ax1)
    >>> ax1.set_title('original image and clipper')
    >>> dst_raster.plot(ax=ax2)
    >>> ax2.set_title('clipped image')
    >>> plt.show()
    """
    assert src_raster.geo_transform is not None, "src_raster.geo_transform should not be None"
    src_ds = src_raster.to_gdal_ds()
    dst_ds = gdal.Warp('', src_ds, format= 'MEM', outputBounds=extent, cropToCutline=True)
    dst_raster = tgp.read_gdal_ds(dst_ds)
    return dst_raster

def zonal_stats(src_poly, src_raster, operator=['mean']):
    """Calculate the statistic value for each zone defined by src_poly, base on values from src_raster. 

    Parameters
    ----------
    src_poly: Geopandas.GeoDataFrame
        The Zone dataset to calculated statistic values.
    src_raster: Raster
        Which value dataset to be calculated statistic values.
    operator: list of {'mean', 'max', 'min', 'median', 'std', 'sum', 'count'}, optional, defalut: ['mean']
        The statistic operator to be used.

    Returns
    -------
    vector: Geopandas.GeoDataFrame 
        Vectorize result.

    Examples
    -------- 
    >>> import geopandas as gpd
    >>> import TronGisPy as tgp
    >>> from TronGisPy import ShapeGrid
    >>> src_raster_fp = tgp.get_testing_fp('satellite_tif')
    >>> src_poly_fp = tgp.get_testing_fp('satellite_tif_clipper')
    >>> src_raster = tgp.read_raster(src_raster_fp)
    >>> src_poly = gpd.read_file(src_poly_fp)
    >>> df_shp = ShapeGrid.zonal_stats(src_poly, src_raster, operator=['mean'])
    >>> df_shp.head()
    """
    assert src_raster.geo_transform is not None, "src_raster.geo_transform should not be None"
    assert isinstance(operator, list), "operator should be a list of string. ex: ['mean']"
    df_shp = src_poly.copy()
    df_shp['poly_idx'] = list(range(len(df_shp)))
    df_shp['poly_idx'] = df_shp['poly_idx'].astype('float')
    poly_rst = tgp.ShapeGrid.rasterize_layer(df_shp, src_raster.rows, src_raster.cols, src_raster.geo_transform, 'poly_idx', all_touched=True, no_data_value=np.nan)
    X_combine = np.concatenate([poly_rst.data, src_raster.data], axis=-1)
    X_combine_df = pd.DataFrame(X_combine.reshape(-1, 2))
    X_groupby = X_combine_df.groupby(0, as_index=False)
    for op in operator:
        if op == 'mean':
            df_shp = df_shp.merge(X_groupby.mean().rename(columns={0:'poly_idx', 1:f'zonal_{op}'}), on='poly_idx', how='left')
        elif op == 'max':
            df_shp = df_shp.merge(X_groupby.max().rename(columns={0:'poly_idx', 1:f'zonal_{op}'}), on='poly_idx', how='left')
        elif op == 'min':
            df_shp = df_shp.merge(X_groupby.min().rename(columns={0:'poly_idx', 1:f'zonal_{op}'}), on='poly_idx', how='left')
        elif op == 'median':
            df_shp = df_shp.merge(X_groupby.median().rename(columns={0:'poly_idx', 1:f'zonal_{op}'}), on='poly_idx', how='left')
        elif op == 'sum':
            df_shp = df_shp.merge(X_groupby.sum().rename(columns={0:'poly_idx', 1:f'zonal_{op}'}), on='poly_idx', how='left')
        elif op == 'std':
            df_shp = df_shp.merge(X_groupby.std().rename(columns={0:'poly_idx', 1:f'zonal_{op}'}), on='poly_idx', how='left')
        elif op == 'count':
            df_shp = df_shp.merge(X_groupby.count().rename(columns={0:'poly_idx', 1:f'zonal_{op}'}), on='poly_idx', how='left')
        else:
            assert False, "no this operator"
    return df_shp


import numpy as np

class Normalizer():
    """Normalize the digital values of the image into 0~1 range."""

    def clip_by_percentage(self, data, clip_percentage=(0.02, 0.98)):
        data_temp = data.copy()
        data_temp_masked = data_temp[~np.isnan(data_temp)]
        idx_st = int(len(data_temp_masked) * clip_percentage[0])
        idx_end = int((len(data_temp_masked) * clip_percentage[1]) - 10**-6)
        X_sorted = np.sort(data_temp_masked)
        data_min = X_sorted[idx_st]
        data_max = X_sorted[idx_end]
        data_temp_masked[data_temp_masked<data_min] = data_min
        data_temp_masked[data_temp_masked>data_max] = data_max
        data_temp[~np.isnan(data_temp)] = data_temp_masked
        return data_temp

    def clip_by_min_max(self, data, min_max=(0, 100)):
        data_temp = data.copy()
        data_min, data_max = min_max
        data_temp[data_temp<data_min] = data_min
        data_temp[data_temp>data_max] = data_max
        return data_temp

    def fit(self, X, min_max_val=None):
        """find the min and max for the normalizer.

        Parameters
        ----------
        X: array_like
            The image used to normalize.
        min_max_val: tuple, optional
            The min and max limitation on the normalizer e.g. (50, 150).
        """
        if min_max_val is not None:
            self.min, self.max = min_max_val
        else:
            self.min, self.max = np.nanmin(X), np.nanmax(X)
        assert self.min != self.max, "The image has only single value, that cannot be normed!"

    def transform(self, X):
        """Normalize the image.

        Parameters
        ----------
        X: array_like
            The image used to normalize.

        Returns
        -------
        Y: ndarray
            The normalized image.
        """
        X = (X - self.min) / (self.max - self.min)
        X_not_nan = X[~np.isnan(X)]
        X_not_nan[(X_not_nan<0)] = 0
        X_not_nan[(X_not_nan>1)] = 1
        return X

    def fit_transform(self, X, min_max_val=None, clip_percentage=None, clip_min_max=None):
        """Combine the fit and transform funtion.

        Parameters
        -------
        X: ndarray
            The normalized image.
        min_max_val: tuple, optional
            The min and max limitation on the normalizer e.g. (50, 150).
        clip_percentage: tuple, optional
            give a two element tuple represents the start and end percentage of 
            the data to cut in head and tail e.g. (0.02, 0.98)
        clip_min_max: tuple, optional
            give a two element tuple represents the min and max to cut in head 
            and tail e.g. (0, 100)

        Returns
        -------
        Y: ndarray
            The normalized image.
        """
        if clip_percentage is not None:
            assert not (clip_percentage is not None and clip_min_max is not None), "should not set clip_percentage and clip_min_max at the same time!"
            assert len(clip_percentage) == 2, "clip_percentage two element tuple"
            X = self.clip_by_percentage(X, clip_percentage=clip_percentage)
        elif clip_min_max is not None:
            assert len(clip_min_max) == 2, "clip_min_max two element tuple"
            X = self.clip_by_min_max(X, min_max=clip_min_max)
        self.fit(X, min_max_val=min_max_val)
        return self.transform(X)
        
    def reverse_transform(self, Y):
        """Reverse the transform funtion of te normalizer.

        Parameters
        ----------
        Y: ndarray
            The normalized image.

        Returns
        -------
        X: ndarray
            The un-normalized image.
        """
        return (Y * (self.max - self.min)) + self.min



import osr
import gdal
import numba
import affine
import pyproj
import numpy as np

@numba.jit(nopython=True)
def __affine_from_gdal(c, a, b, f, d, e):
    members = [a, b, c, d, e, f, 0.0, 0.0, 1.0]
    mat3x3 = [x * 1.0 for x in members[:-3]]
    return mat3x3

@numba.jit(nopython=True)
def __invert_geo_transform( a, b, c, d, e, f):
    determinant = a * e - b * d
    idet = 1.0 / determinant
    sa, sb, sc, sd, se, sf = a, b, c, d, e, f
    ra = se * idet
    rb = -sb * idet
    rd = -sd * idet
    re = sa * idet
    return [ra, rb, -sc * ra - sf * rb,
         rd, re, -sc * rd - sf * re,
         0.0, 0.0, 1.0]

@numba.jit(nopython=True)
def __numba_transfer_coord_to_xy(x, y, a, b, c, d, e, f):
    coord_x, coord_y = x, y
    geo_transform = (a, b, c, d, e, f)
    forward_transform =  __affine_from_gdal(*geo_transform)
    reverse_transform = __invert_geo_transform(forward_transform[0], forward_transform[1], forward_transform[2],
                                             forward_transform[3], forward_transform[4], forward_transform[5] )
    reverse_transform = np.array(reverse_transform).reshape((3, 3))
    x, y, _ = reverse_transform.dot(np.array([coord_x, coord_y, 1]))
    # x, y = np.int(x), np.int(y)
    return x, y

@numba.jit(nopython=True)
def coords_to_npidxs(coords, geo_transform):
    """Find the cells' numpy index the coordinates belong to using 
    the following functions.

    | coord_lng |   | a  b  c | | npidx_col |
    | coord_lat | = | d  e  f | | npidx_row |
    |     1     |   | 0  0  1 | |     1     |

    Parameters
    ----------
    coords: array_like
        The coordinates with shape (n_points, 2). The order of last dimension is (lng, lat).
    geo_transform: tuple or list
        Affine transform parameters (c, a, b, f, d, e = geo_transform).

    Returns
    -------
    npidxs: ndarray
        The numpy indeices with shape (n_points, 2). The order of last dimension is (row_idx, col_idx).

    Examples
    --------
    >>> import numpy as np
    >>> import TronGisPy as tgp
    >>> geo_transform = tgp.get_raster_info(tgp.get_testing_fp(), 'geo_transform')
    >>> coords = np.array([(271986.23416588, 2769971.94666942)])
    >>> tgp.coords_to_npidxs(coords, geo_transform)
    array([[1, 3]], dtype=int64)
    """
    group_npidx = []
    for i in range(0, len(coords)):
        x, y = coords[i][0], coords[i][1]
        x, y = __numba_transfer_coord_to_xy(x, y, *geo_transform)
        group_npidx.append((y, x))
    return np.array(group_npidx, np.int64)

def npidxs_to_coords(npidxs, geo_transform): # TODO: reproduce to numba
    """Get coordinates of cells' left-top corner by its numpy index using 
    the following functions.

    | coord_lng |   | a  b  c | | npidx_col |
    | coord_lat | = | d  e  f | | npidx_row |
    |     1     |   | 0  0  1 | |     1     |

    Parameters
    ----------
    npidxs: array_like
        The numpy indeices with shape (n_points, 2). The order of last dimension is (row_idx, col_idx).
    geo_transform: tuple or list
        Affine transform parameters (c, a, b, f, d, e = geo_transform).

    Returns
    -------
    coords: ndarray 
        The coordinates with shape (n_points, 2). The order of last dimension is (lng, lat).

    Examples
    --------
    >>> import TronGisPy as tgp
    >>> geo_transform = tgp.get_raster_info(tgp.get_testing_fp(), 'geo_transform')
    >>> tgp.npidxs_to_coords([(1,3)], geo_transform)
    array([[ 271986.23416588, 2769971.94666942]])
    """
    npidxs = np.array(npidxs)
    assert len(npidxs.shape) == 2, "npidxs.shape should be (n_points, 2)"

    # prepare M
    c, a, b, f, d, e = geo_transform
    M = np.array([[a, b, c], 
                  [d, e, f], 
                  [0, 0, 1]])

    # prepare npidxs_maxtrix
    row_idxs, col_idxs = np.array(npidxs).T
    npidxs_maxtrix = np.ones((len(npidxs), 3)).T # => (3, -1)
    npidxs_maxtrix[0], npidxs_maxtrix[1] = col_idxs, row_idxs

    # apply multiplication
    coords = np.matmul(M, npidxs_maxtrix).T[:, :2] # (3, 3) ‧ (3, -1) => (3, -1) => (-1, 3)
    return coords

def npidxs_to_coord_polygons(npidxs, geo_transform):
    """Get coordinates of cells' four corners by its numpy index. 

    Parameters
    ----------
    npidxs: array_like
        The numpy indeices with shape (n_points, 2). The order of last dimension is (row_idx, col_idx).
    geo_transform: tuple or list 
        Affine transform parameters (c, a, b, f, d, e = geo_transform).

    Returns
    -------
    poly_points: ndarray
        The four corner coordinates for each npidxs with shape (-1, 4, 2). The order of last dimension is (lng, lat).

    Examples
    --------
    >>> import TronGisPy as tgp
    >>> geo_transform = tgp.get_raster_info(tgp.get_testing_fp(), 'geo_transform')
    >>> tgp.npidxs_to_coord_polygons([(1,3)], geo_transform)
    array([[[ 271986.23416588, 2769971.94666942],
            [ 271987.35278783, 2769971.94666942],
            [ 271987.35278783, 2769970.82803885],
            [ 271986.23416588, 2769970.82803885]]])
    """
    npidxs = np.array(npidxs)
    assert len(npidxs.shape) == 2, "npidxs.shape should be (n_points, 2)"
    poly_points = [ [npidxs[:, 0]+0, npidxs[:, 1]+0],  # ul
                    [npidxs[:, 0]+0, npidxs[:, 1]+1],  # ur
                    [npidxs[:, 0]+1, npidxs[:, 1]+1],  # lr
                    [npidxs[:, 0]+1, npidxs[:, 1]+0],] # ll
    poly_points = np.array(poly_points).transpose([2, 0, 1]).reshape(-1, 2)
    poly_points = npidxs_to_coords(poly_points, geo_transform)
    poly_points = poly_points.reshape(-1, 4, 2)
    return poly_points

def get_extent(rows, cols, geo_transform, return_type='poly'):
    """Get the boundary of a raster file

    Parameters
    ----------
    rows: int
        The number of rows in the raster.
    cols: int
        The number of cols in the raster.
    geo_transform : tuple or list
        Affine transform parameters (c, a, b, f, d, e = geo_transform).
    return_type: {'poly', 'plot', 'gdal'}
        If 'poly', return four corner coordinates. If plot, return (xmin, xmax, ymin, ymax). If 'gdal', return (xmin, ymin, xmax, ymax). 

    Returns
    -------
    extent: ndarray or tuple
        Depends on return_type. If 'poly', return four corner coordinates. If plot, return (xmin, xmax, ymin, ymax). If 'gdal', return (xmin, ymin, xmax, ymax). 

    Examples
    --------
    >>> import TronGisPy as tgp
    >>> rows, cols, geo_transform = tgp.get_raster_info(tgp.get_testing_fp(), ['rows', 'cols', 'geo_transform'])
    >>> tgp.get_extent(rows, cols, geo_transform, return_type='plot')
    (271982.8783, 272736.8295, 2769215.7524, 2769973.0653)
    """
    assert return_type in ['poly', 'plot', 'gdal'], "return_type should be in ['poly', 'plot', 'gdal']"
    points = [[0,0], [0,cols], [rows,cols], [rows,0]]
    poly = npidxs_to_coords(points, geo_transform)
    if return_type == 'poly':
        extent = poly
    elif return_type == 'plot':
        extent = (np.min(poly[:, 0]), np.max(poly[:, 0]), np.min(poly[:, 1]), np.max(poly[:, 1]))
    elif return_type == 'gdal':
        extent = (np.min(poly[:, 0]), np.min(poly[:, 1]), np.max(poly[:, 0]), np.max(poly[:, 1]))
    return extent

def epsg_to_wkt(epsg=4326):
    """convert epsg code to well known text (WKT).

    Parameters
    ----------
    epsg: int, optional
        The epsg code.

    Returns
    -------
    wkt: str
        The well known text of the epsg code.
    
    Examples
    --------
    >>> import TronGisPy as tgp
    >>> tgp.epsg_to_wkt(4326)
    'GEOGCRS["WGS 84",DATUM["World Geodetic System 1984",ELLIPSOID["WGS 84",6378137,298.257223563,LENGTHUNIT["metre",1]]],PRIMEM["Greenwich",0,ANGLEUNIT["degree",0.0174532925199433]],CS[ellipsoidal,2],AXIS["geodetic latitude (Lat)",north,ORDER[1],ANGLEUNIT["degree",0.0174532925199433]],AXIS["geodetic longitude (Lon)",east,ORDER[2],ANGLEUNIT["degree",0.0174532925199433]],USAGE[SCOPE["unknown"],AREA["World"],BBOX[-90,-180,90,180]],ID["EPSG",4326]]'
    """
    return pyproj.CRS.from_epsg(epsg).to_wkt()

def wkt_to_epsg(wkt, pkg='both'):
    """convert well known text (WKT) to epsg code.

    Parameters
    ----------
    wkt: str 
        The well known text of the epsg code.
    pkg: {'pyproj', 'osr', 'both'}, optional, default: both
        Which package to use to convert the wkt. Should be in {'pyproj', 'osr', 'both'}.

    Returns
    -------
    epsg: int
        The epsg code.

    Examples
    --------
    >>> import TronGisPy as tgp
    >>> projection = tgp.get_raster_info(tgp.get_testing_fp(), 'projection')
    >>> tgp.wkt_to_epsg(projection)
    """
    if pkg == 'osr':
        srs = osr.SpatialReference(wkt=wkt)
        epsg = srs.GetAuthorityCode(None)
    elif pkg == 'pyproj':
        epsg = pyproj.CRS(wkt).to_epsg()
    elif pkg == 'both':
        srs = osr.SpatialReference(wkt=wkt) 
        epsg = srs.GetAuthorityCode(None)
        if epsg is None:
            epsg = pyproj.CRS(wkt).to_epsg()

    if epsg is not None:
        return int(epsg)
    else:
        assert False, "the wkt connot be converted."



import os
import gdal
import time
import shutil
import numpy as np
import TronGisPy as tgp
from TronGisPy import ShapeGrid

# operation on raster files
# ===========================
def get_raster_info(fp, attributes=["rows", "cols", "bands", "geo_transform", "projection", "gdaldtype", "no_data_value", "metadata"]):
    """Get the geo-information of the raster file including rows, cols, bands, geo_transform, 
    projection, gdaldtype, no_data_value, metadata attributes.

    Parameters
    ----------
    fp: str
        File path of the raster file.
    attributes: list or str
        Which attributes to get e.g. ["rows", "cols", "bands", "geo_transform"].

    Returns
    -------
    geo_info : list or attribute
        Attributes of the raster.

    Examples
    --------
    >>> import TronGisPy as tgp
    >>> raster_fp = tgp.get_testing_fp()
    >>> rows, cols, bands, geo_transform, projection, gdaldtype, no_data_value, metadata = tgp.get_raster_info(raster_fp)
    >>> rows, cols, bands
    (677, 674, 3)
    >>> tgp.get_raster_info(raster_fp, 'projection')
    'PROJCS["TWD97 / TM2 zone 121",GEOGCS["TWD97",DATUM["Taiwan_Datum_1997",SPHEROID["GRS 1980",6378137,298.257222101,AUTHORITY["EPSG","7019"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","1026"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","3824"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",121],PARAMETER["scale_factor",0.9999],PARAMETER["false_easting",250000],PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","3826"]]'
    >>> tgp.get_raster_info(raster_fp, ['rows', 'cols'])
    [677, 674]
    """
    ds = gdal.Open(fp)
    rows, cols, bands = ds.RasterYSize, ds.RasterXSize, ds.RasterCount
    geo_transform, projection, metadata = ds.GetGeoTransform(), ds.GetProjection(), ds.GetMetadata()
    gdaldtype = ds.GetRasterBand(1).DataType
    no_data_value = ds.GetRasterBand(1).GetNoDataValue()
    ds = None 

    geo_info = rows, cols, bands, geo_transform, projection, gdaldtype, no_data_value, metadata
    default_attributes = ["rows", "cols", "bands", "geo_transform", "projection", "gdaldtype", "no_data_value", "metadata"]
    if type(attributes) == list:
        attributes_idxs = [default_attributes.index(attribute) for attribute in attributes]
        return [geo_info[attributes_idx] for attributes_idx in attributes_idxs]
    else:
        attributes_idx = default_attributes.index(attributes)
        return geo_info[attributes_idx]

def get_raster_data(fp):
    """Get the digital number of the raster file.

    Parameters
    ----------
    fp: str
        File path of the raster file.

    Returns
    -------
    data : ndarray
        Digital number of the raster file whose shape will be (rows, cols, bnads).

    Examples
    --------
    >>> import TronGisPy as tgp 
    >>> raster_fp = tgp.get_testing_fp()
    >>> data = tgp.get_raster_data(raster_fp)
    >>> type(data) 
    <class 'numpy.ndarray'>
    >>> data.shape 
    (677, 674, 3)
    """
    ds = gdal.Open(fp)
    data = ds.ReadAsArray()
    ds = None 
    if len(data.shape) == 2:
        data = np.expand_dims(data, axis=0)
    return np.transpose(data, axes=[1,2,0])

def get_raster_extent(fp, return_type='poly'):
    """Get the boundary of the raster file.

    Parameters
    ----------
    fp: str
        File path of the raster file.
    return_type: {'poly', 'plot', 'gdal'}
        If 'poly', return four corner coordinates. If plot, return (xmin, xmax, ymin, ymax). If 'gdal', return (xmin, ymin, xmax, ymax). 

    Returns
    -------
    extent: ndarray or tuple
        Depends on return_type. If 'poly', return four corner coordinates. If plot, return (xmin, xmax, ymin, ymax). If 'gdal', return (xmin, ymin, xmax, ymax). 

    Examples
    --------
    >>> import TronGisPy as tgp 
    >>> from shapely.geometry import Polygon
    >>> raster_fp = tgp.get_testing_fp() 
    >>> extent = tgp.get_raster_extent(raster_fp) 
    >>> Polygon(extent).area
    570976.9697303267
    """
    rows, cols, geo_transform = get_raster_info(fp, ['rows', 'cols', 'geo_transform'])
    return tgp.get_extent(rows, cols, geo_transform, return_type)

def update_raster_info(fp, geo_transform=None, projection=None, gdaldtype=None, no_data_value=None, metadata=None):
    """Update the geo-information of the raster file including geo_transform, projection, gdaldtype, 
    no_data_value, metadata attributes.

    Parameters
    ----------
    fp: str
        File path of the raster file.
    geo_transform: tuple or list
        Affine transform parameters (c, a, b, f, d, e = geo_transform). 
    projection: str, optional
        The well known text (WKT) of the raster which can be generated from 
        `TronGisPy.epsg_to_wkt(<epsg_code>)`
    gdaldtype: int, optional
        The type of the cell defined in gdal which will affect the information 
        to be stored when saving the file. This can be generate from `gdal.GDT_XXX` 
        such as `gdal.GDT_Int32` equals 5 and `gdal.GDT_Float32` equals 6.
    no_data_value: int or float, optional
        Define which value to replace nan in numpy array when saving a raster file.
    metadata: dict, optional
        Define the metadata of the raster file.
    """
    all_none = geo_transform is None and projection is None and gdaldtype is None and no_data_value is None and metadata is None
    assert not all_none, "at least one of geo_transform and projection params should not be None!"
    ds = gdal.Open(fp, gdal.GA_Update)
    if geo_transform is not None:
        ds.SetGeoTransform(geo_transform)
    if projection is not None:
        ds.SetProjection(projection)
    if metadata is not None:
        ds.SetMetadata(metadata)
    if no_data_value is not None:
        for b in range(ds.RasterCount):
            band = ds.GetRasterBand(b+1)
            band.SetNoDataValue(no_data_value)
            band.FlushCache()
    ds = None 


# operation on Raster class
# ===========================
def read_raster(fp, fill_na=False):
    """Read raster file as `TronGisPy.Raster` object.

    Parameters
    ----------
    fp: str 
        File path of the raster file.

    Returns
    -------
    raster: Raster
        output raster.

    Examples
    --------
    >>> import TronGisPy as tgp 
    >>> from shapely.geometry import Polygon
    >>> raster_fp = tgp.get_testing_fp() 
    >>> raster = tgp.read_raster(raster_fp) 
    >>> raster
    shape: (677, 674, 3)
    geo_transform: (271982.8783, 1.1186219584569888, 0.0, 2769973.0653, 0.0, -1.1186305760705852)
    projection: PROJCS["TWD97 / TM2 zone 121",GEOGCS["TWD97",DATUM["Taiwan_Datum_1997",SPHEROID["GRS 1980",6378137,298.257222101,AUTHORITY["EPSG","7019"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","1026"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","3824"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",121],PARAMETER["scale_factor",0.9999],PARAMETER["false_easting",250000],PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","3826"]]
    no_data_value: -32768.0
    metadata: {'AREA_OR_POINT': 'Area'}
    """
    from TronGisPy import Raster
    data = get_raster_data(fp)
    rows, cols, bands, geo_transform, projection, gdaldtype, no_data_value, metadata = get_raster_info(fp)
    raster = Raster(data, geo_transform, projection, gdaldtype, no_data_value, metadata)
    if fill_na and np.sum(np.isnan(raster.data)):
        raster.fill_na()
    return raster

def write_raster(fp, data, geo_transform=None, projection=None, gdaldtype=None, no_data_value=None, metadata=None):
    """Write raster file.

    Parameters
    ----------
    fp: str
        File path of the raster file.
    geo_transform: tuple or list, optional
        Affine transform parameters (c, a, b, f, d, e = geo_transform). 
    projection: str, optional
        The well known text (WKT) of the raster which can be generate from `TronGisPy.epsg_to_wkt(<epsg_code>)`
    gdaldtype: int, optional
        The type of the cell defined in gdal which will affect the information 
        to be stored when saving the file. This can be generate from `gdal.GDT_XXX` 
        such as `gdal.GDT_Int32` equals 5 and `gdal.GDT_Float32` equals 6.
    no_data_value: int or float, optional
        Define which value to replace nan in numpy array when saving a raster file.
    metadata: dict, optional
        Define the metadata of the raster file.
    """
    if len(data.shape) == 2:
        data = np.expand_dims(data, axis=2)
    rows, cols, bands = data.shape
    gdaldtype = tgp.npdtype_to_gdaldtype(data.dtype) if gdaldtype is None else gdaldtype
    ds = gdal.GetDriverByName('GTiff').Create(fp, cols, rows, bands, gdaldtype) # dst_filename, xsize=512, ysize=512, bands=1, eType=gdal.GDT_Byte
    if geo_transform is not None:
        ds.SetGeoTransform(geo_transform)
    if projection is not None:
        ds.SetProjection(projection)
    if metadata is not None:
        ds.SetMetadata(metadata)

    for b in range(bands):
        band = ds.GetRasterBand(b+1)
        band.WriteArray(data[:, :, b], 0, 0)
        if no_data_value is not None:
            band.SetNoDataValue(no_data_value)
        band.FlushCache()
    ds = None

def read_gdal_ds(ds):
    """Read gdal DataSource as `TronGisPy.Raster` object.

    Parameters
    ----------
    ds: gdal.DataSource.

    Returns
    -------
    raster: Raster.
        output raster.
    """
    rows, cols, bands = ds.RasterYSize, ds.RasterXSize, ds.RasterCount
    geo_transform, projection, metadata = ds.GetGeoTransform(), ds.GetProjection(), ds.GetMetadata()
    gdaldtype = ds.GetRasterBand(1).DataType
    no_data_value = ds.GetRasterBand(1).GetNoDataValue()
    data = ds.ReadAsArray()
    if len(data.shape) == 2:
        data = np.expand_dims(data, axis=0)
    data = np.transpose(data, axes=[1,2,0])
    from TronGisPy import Raster
    return Raster(data, geo_transform, projection, gdaldtype, no_data_value, metadata)


def write_gdal_ds(data=None, bands=None, cols=None, rows=None, geo_transform=None, projection=None, gdaldtype=None, no_data_value=None, metadata=None):
    """Build the gdal DataSource from geo-information attributes.

    Parameters
    ----------
    data: array_like, optional
        The digital number for each cell of the raster.  Data is in 
        (n_rows, n_cols, n_bands) shape.
    bands: int, optional
        Number of bands.
    cols: int, optional
        Number of cols.
    rows: int, optional
        Number of rows.
    geo_transform: tuple or list, optional
        Affine transform parameters (c, a, b, f, d, e = geo_transform). 
    projection: str, optional
        The well known text (WKT) of the raster which can be generate 
        from `TronGisPy.epsg_to_wkt(<epsg_code>)`
    gdaldtype: int, optional
        The type of the cell defined in gdal which will affect the information 
        to be stored when saving the file. This can be generate from `gdal.GDT_XXX` 
        such as `gdal.GDT_Int32` equals 5 and `gdal.GDT_Float32` equals 6.
    no_data_value: int or float, optional
        Define which value to replace nan in numpy array when saving a raster file.
    metadata: dict, optional
        Define the metadata of the raster file.

    Returns
    -------
    raster: Raster.
        output raster.

    Examples
    --------
    >>> import TronGisPy as tgp 
    >>> raster_fp = tgp.get_testing_fp() 
    >>> data = tgp.get_raster_data(raster_fp)
    >>> geo_transform, projection, gdaldtype, no_data_value = tgp.get_raster_info(raster_fp, ["geo_transform", "projection", "gdaldtype", "no_data_value"])
    >>> ds = tgp.write_gdal_ds(data, geo_transform=geo_transform, projection=projection, gdaldtype=gdaldtype, no_data_value=no_data_value)
    >>> raster = tgp.read_gdal_ds(ds)
    >>> raster
    shape: (677, 674, 3)
    geo_transform: (271982.8783, 1.1186219584569888, 0.0, 2769973.0653, 0.0, -1.1186305760705852)
    projection: PROJCS["TWD97 / TM2 zone 121",GEOGCS["TWD97",DATUM["Taiwan_Datum_1997",SPHEROID["GRS 1980",6378137,298.257222101,AUTHORITY["EPSG","7019"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","1026"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","3824"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",121],PARAMETER["scale_factor",0.9999],PARAMETER["false_easting",250000],PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","3826"]]
    no_data_value: -32768.0
    metadata: {}
    """
    if data is None:
        assert (bands is not None) and (cols is not None) and (rows is not None), "bands, cols, rows should not be None"
        assert (gdaldtype is not None), "gdaldtype should not be None"
    else:    
        if len(data.shape) == 2:
            data = np.expand_dims(data, axis=2)
        bands = data.shape[2] if bands is None else bands
        cols = data.shape[1] if cols is None else cols
        rows = data.shape[0] if rows is None else rows
        gdaldtype = tgp.npdtype_to_gdaldtype(data.dtype) if gdaldtype is None else gdaldtype

    ds = gdal.GetDriverByName('MEM').Create('', cols, rows, bands, gdaldtype) # dst_filename, xsize=512, ysize=512, bands=1, eType=gdal.GDT_Byte
    if geo_transform is not None:
        ds.SetGeoTransform(geo_transform)
    if projection is not None:
        ds.SetProjection(projection)
    if metadata is not None:
        ds.SetMetadata(metadata)

    if no_data_value is not None:
        for b in range(bands):
            band = ds.GetRasterBand(b+1)
            band.SetNoDataValue(no_data_value)
            band.WriteArray(np.full((rows, cols), no_data_value), 0, 0)
            band.FlushCache()
    
    if data is not None:
        for b in range(data.shape[2]):
            band = ds.GetRasterBand(b+1)
            band.WriteArray(data[:, :, b], 0, 0)
            band.FlushCache()
    return ds

def remove_shp(shp_fp):
    """Remove shapefile and all its related files.

    Parameters
    ----------
    shp_fp: str 
        File path of the shapefile.
    """
    assert shp_fp.endswith('.shp'), 'shp_fp should be ends with ".shp"'
    dst_shp_fp = os.path.abspath(shp_fp)
    base_dir = os.path.split(dst_shp_fp)[0]
    shp_fn = os.path.split(dst_shp_fp)[-1].split(".")[0]
    del_fps = [os.path.join(base_dir, f) for f in  os.listdir(base_dir) if shp_fn == f.split('.')[0]]
    for fp in del_fps:
        os.remove(fp)

# testing data
# ===========================
def get_testing_fp(fn=None):
    """Get the testing file built-in TronGisPy.

    Parameters
    ----------
    fn: str
        Choice candidates includes satellite_tif, satellite_tif_clipper,
        satellite_tif_kmeans, rasterized_image, rasterized_image_1,
        poly_to_be_clipped, point_to_be_clipped, line_to_be_clipped,
        multiline_to_be_clipped, shp_clipper, remap_rgb_clipper_path,
        remap_ndvi_path, dem_process_path, tif_forinterpolation,
        aero_triangulation_PXYZs

    Returns
    -------
    fp: str
        The path of testing file.

    Examples
    --------
    >>> import TronGisPy as tgp 
    >>> raster_fp = tgp.get_testing_fp() 
    >>> raster = tgp.read_raster(raster_fp)
    >>> raster.plot()
    """
    base_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(base_dir, 'data')
    if fn is None:
        fp = os.path.join(data_dir, 'remap', 'rgb_3826_clipper.tif')
    elif fn == 'satellite_tif':
        fp = os.path.join(data_dir, 'satellite_tif', 'satellite_tif.tif')
    elif fn == 'satellite_tif_clipper':
        fp = os.path.join(data_dir, 'satellite_tif_clipper', 'satellite_tif_clipper.shp')
    elif fn == 'satellite_tif_kmeans':
        fp = os.path.join(data_dir, 'satellite_tif_kmeans', 'satellite_tif_kmeans.tif')
    elif fn == 'rasterized_image':
        fp = os.path.join(data_dir, 'rasterized_image', 'rasterized_image.tif')
    elif fn == 'rasterized_image_1':
        fp = os.path.join(data_dir, 'rasterized_image', 'rasterized_image_1.tif')
    elif fn == 'poly_to_be_clipped':
        fp = os.path.join(data_dir, 'clip_shp_by_shp', 'poly_to_be_clipped.shp')
    elif fn == 'point_to_be_clipped':
        fp = os.path.join(data_dir, 'clip_shp_by_shp', 'point_to_be_clipped.shp')
    elif fn == 'line_to_be_clipped':
        fp = os.path.join(data_dir, 'clip_shp_by_shp', 'line_to_be_clipped.shp')
    elif fn == 'multiline_to_be_clipped':
        fp = os.path.join(data_dir, 'clip_shp_by_shp', 'multiline_to_be_clipped.shp')
    elif fn == 'shp_clipper':
        fp = os.path.join(data_dir, 'clip_shp_by_shp', 'shp_clipper.shp')
    elif fn == 'remap_rgb_clipper_path':
        fp = os.path.join(data_dir, 'remap', 'rgb_3826_clipper.tif')
    elif fn == 'remap_ndvi_path':
        fp = os.path.join(data_dir, 'remap', 'ndvi_32651.tif')
    elif fn == 'dem_process_path':
        fp = os.path.join(data_dir, 'dem_tif', 'crop_dem.tif')
    elif fn == 'tif_forinterpolation':
        fp = os.path.join(data_dir, 'interpolation', 'X_forinterpolation.tif')
    elif fn == 'aero_triangulation_PXYZs':
        fp = os.path.join(data_dir, 'aero_triangulation', 'P_XYZs.npy')
    elif fn == 'flipped_gt':
        fp = os.path.join(data_dir, 'flipped_gt', 'flipped_gt.tif')
    elif fn == 'rotate_tif':
        fp = os.path.join(data_dir, 'rotate_tif', 'rotate_tif.tif')
    elif fn == 'multiple_poly_clipper':
        fp = os.path.join(data_dir, 'multiple_poly_clipping', 'df_farm_clipped.shp')
    elif fn == 'multiple_poly_clip_ras':
        fp = os.path.join(data_dir, 'multiple_poly_clipping', 'ras_sat_clipped.tif')
    elif fn == 'norm':
        fp = os.path.join(data_dir, 'norm', '100131i_39_0049_1527_052_019.tif')
    else:
        assert False, "cannot find the file!!"
    return os.path.abspath(fp)


# def create_temp_dir():
#     base_dir = os.path.dirname(os.path.realpath(__file__))
#     temp_dir = os.path.abspath(os.path.join(base_dir, 'temp'))
#     if not os.path.isdir(temp_dir):
#         os.mkdir(temp_dir)
#     else:
#         shutil.rmtree(temp_dir)
#         time.sleep(0.5)
#         os.mkdir(temp_dir)
#     return temp_dir

# def remove_temp_dir():
#     base_dir = os.path.dirname(os.path.realpath(__file__))
#     temp_dir = os.path.abspath(os.path.join(base_dir, 'temp'))
#     shutil.rmtree(temp_dir)
#     return temp_dir


def create_temp_dir_when_not_exists():
    base_dir = os.path.dirname(os.path.realpath(__file__))
    temp_dir = os.path.abspath(os.path.join(base_dir, 'temp'))
    if not os.path.isdir(temp_dir):
        os.mkdir(temp_dir)
    return temp_dir




import os
import gdal
import numpy as np
from numba import jit
import TronGisPy as tgp
from scipy.interpolate import griddata


def __band_interpolation(data, method='linear', no_data_value=None):
    """be used in img_interpolation"""
    assert len(data.shape) == 2, "data should be one band image!"
    if no_data_value is not None:
        data[data==no_data_value] = np.nan
    points = np.array(np.where(~np.isnan(data))).T
    values = data[~np.isnan(data)].copy()
    grid_x, grid_y = np.where(np.ones_like(data))
    grid_x, grid_y = grid_x.reshape(data.shape), grid_y.reshape(data.shape)
    X_interp = griddata(points, values, (grid_x, grid_y), method=method)
    return X_interp

def img_interpolation(data, method='linear', no_data_value=None):
    """Interpolate values on specific cells (generally nan cell) of the image data. 
    use scipy.interpolate.griddata engine. Note: Use majority_interpolation, mean_interpolation
    and gdal_fillnodata to speed up.

    Parameters
    ----------
    data: array_like
        The digital number of the image which is in (n_rows, n_cols, n_bands) shape.
    method: {'nearest', 'linear', 'cubic'}, optional, default: linear
        use scipy.interpolate.griddata function.
    no_data_value: int or float, optional, default: None
        The vealue to be filled with interpolated value. If no_data_value == None, use np.nan as no_data_value.

    Returns
    -------
    data_interp: ndarray. 
        Interpolation result.

    Examples
    -------- 
    >>> from matplotlib import pyplot as plt
    >>> import TronGisPy as tgp 
    >>> from TronGisPy import Interpolation
    >>> raster_fp = tgp.get_testing_fp('tif_forinterpolation')
    >>> raster = tgp.read_raster(raster_fp)
    >>> data_interp = Interpolation.img_interpolation(raster.data)
    >>> fig, (ax1, ax2) = plt.subplots(1,2)
    >>> ax1.imshow(raster.data[:, :, 0])
    >>> ax2.imshow(data_interp[:, :, 0])
    >>> plt.show()
    """
    data_interp = data.copy()
    if len(data_interp.shape) == 2:
        data_interp = np.expand_dims(data_interp, axis=2)
    for b in range(data_interp.shape[2]):
        data_interp[:,:,b] = __band_interpolation(data_interp[:,:,b], method='linear', no_data_value=None)
    return data_interp

@jit(nopython=True) # Set "nopython" mode for best performance, equivalent to @njit
def __jit_majority_interpolation(X_stacked, no_data_value): # Function is compiled to machine code when called the first time
    X_interp = np.zeros((X_stacked.shape[0], X_stacked.shape[1]), dtype=np.int32)
    for i in range(X_interp.shape[0]-1):
        for j in range(X_interp.shape[1]-1):
            if X_stacked[i,j][5] == no_data_value: # only calculate majority when value is no_data_value
                vals = X_stacked[i,j][X_stacked[i,j]!=no_data_value]
                v_len = vals.shape[0]
                if v_len !=0: # if at least one value in the convolution is not no_data_value
                    X_interp[i, j] = np.bincount(vals).argmax() # calculate majority 
                else: 
                    X_interp[i, j] = no_data_value
            else:
                X_interp[i, j] = X_stacked[i,j][5]
    return X_interp

@jit(nopython=True) # Set "nopython" mode for best performance, equivalent to @njit
def __jit_mean_interpolation(X_stacked, no_data_value): # Function is compiled to machine code when called the first time
    X_interp = np.zeros((X_stacked.shape[0], X_stacked.shape[1]))
    for i in range(X_interp.shape[0]-1):
        for j in range(X_interp.shape[1]-1):
            if X_stacked[i,j][5] == no_data_value: # only calculate majority when value is no_data_value
                vals = X_stacked[i,j][X_stacked[i,j]!=no_data_value]
                v_len = vals.shape[0]
                if v_len !=0: # if at least one value in the convolution is not no_data_value
                    X_interp[i, j] = np.mean(vals) # calculate mean
                else: 
                    X_interp[i, j] = no_data_value
            else:
                X_interp[i, j] = X_stacked[i,j][5]
    return X_interp

def __generate_stacked_X(X_padded, ws):
    """
    X_stacked = np.stack([
        X_padded[0:-2, 0:-2], X_padded[1:-1, 0:-2], X_padded[2:, 0:-2],
        X_padded[0:-2, 1:-1], X_padded[1:-1, 1:-1], X_padded[2:, 1:-1],
        X_padded[0:-2, 2:  ], X_padded[1:-1, 2:  ], X_padded[2:, 2:  ]
        ] , axis=-1)
    """
    rows, cols = X_padded.shape
    X_stacked = []
    for i in range(ws):
        for j in range(ws):
            X_stacked.append(X_padded[i:rows-(ws-i)+1, j:cols-(ws-j)+1])
    X_stacked = np.stack(X_stacked, axis=-1)
    return X_stacked

def __majority_interpolation_single(X, no_data_value, window_size):
    X_padded = np.pad(X, ((1, 1), (1, 1)), mode='edge')
    X_stacked = __generate_stacked_X(X_padded, window_size)
    X_interp = __jit_majority_interpolation(X_stacked, no_data_value=no_data_value)
    return X_interp

def __mean_interpolation_single(X, no_data_value, window_size):
    X_padded = np.pad(X, ((1, 1), (1, 1)), mode='edge')
    X_stacked = __generate_stacked_X(X_padded, window_size)
    X_interp = __jit_mean_interpolation(X_stacked, no_data_value=no_data_value)
    return X_interp

def majority_interpolation(data, no_data_value=999, window_size=3, loop_to_fill_all=True, loop_limit=5):
    """Interpolate values on specific cells (generally nan cell) using 
    the majority value in the window.

    Parameters
    ----------
    data: array_like
        The digital number of the image which is in (n_rows, n_cols) shape. 
        The dtype of the array should be integer in order to calculate the majority.
    no_data_value: int or float, optional, default: 999
        The value to be filled with interpolated value. If no_data_value == None, 
        use np.nan as no_data_value.
    window_size: int, optional, default: 3
        The size of the window of the convolution to calculate the majority value. 
        Window_size should be odd number.
    loop_to_fill_all: bool, optional, default: True
        Fill all no_data_value until there is no no_data_value value in the data.
    loop_limit: int, optional, default: 5
        The maximum limitation on loop. if loop_to_fill_all==True, loop_limit will 
        be considered. `-1` means no limitation.

    Returns
    -------
    data_interp: ndarray. 
        Interpolation result.

    Examples
    -------- 
    >>> import numpy as np
    >>> import TronGisPy as tgp 
    >>> from TronGisPy import Interpolation
    >>> from matplotlib import pyplot as plt
    >>> raster_fp = tgp.get_testing_fp('tif_forinterpolation')
    >>> raster = tgp.read_raster(raster_fp)
    >>> data = raster.data.copy()[:, :, 0]
    >>> data_interp = data.copy()
    >>> data_interp[np.isnan(data_interp)] = 999
    >>> data_interp = data_interp.astype(np.int)
    >>> data_interp = Interpolation.majority_interpolation(data_interp, no_data_value=999)
    >>> fig, (ax1, ax2) = plt.subplots(1,2)
    >>> ax1.imshow(data)
    >>> ax2.imshow(data_interp)
    >>> plt.show()
    """
    assert len(data.shape) == 2, "data should have onle 2 dimension"
    assert np.issubdtype(data.dtype, np.integer), "data should be in integer type"
    assert window_size%2==1 , "window_size should be odd number"
    data_interp = __majority_interpolation_single(data, no_data_value=no_data_value, window_size=window_size)
    if loop_to_fill_all and (loop_limit != -1):
        loop_count = 0
        while (np.sum(data_interp==no_data_value) > 0) and loop_count<loop_limit:
            data_interp = __majority_interpolation_single(data_interp, no_data_value=no_data_value, window_size=window_size)
            loop_count += 1    
    elif loop_to_fill_all and (loop_limit == -1):
        while np.sum(data_interp==no_data_value) > 0:
            data_interp = __majority_interpolation_single(data_interp, no_data_value=no_data_value, window_size=window_size)
    return data_interp

def mean_interpolation(data, no_data_value=999, window_size=3, loop_to_fill_all=True, loop_limit=5):
    """Interpolate values on specific cells (no_data_value) using 
    the mean value in the window.

    Parameters
    ----------
    data: array_like. 
        The digital number of the image which is in (n_rows, n_cols) shape. 
        The dtype of the array should be integer in order to calculate the majority.
    no_data_value: int or float, optional, default: 999
        The value to be filled with interpolated value. If no_data_value == None, 
        use np.nan as no_data_value.
    window_size: int, optional, default: 3
        The size of the window of the convolution to calculate the majority value. 
        Window_size should be odd number.
    loop_to_fill_all: bool, optional, default: True
        Fill all no_data_value until there is no no_data_value value in the data.
    loop_limit: int, optional, default: 5
        The maximum limitation on loop. if loop_to_fill_all==True, loop_limit will 
        be considered. `-1` means no limitation.

    Returns
    -------
    data_interp: ndarray. 
        Interpolation result.

    Examples
    -------- 
    >>> import numpy as np
    >>> import TronGisPy as tgp 
    >>> from TronGisPy import Interpolation
    >>> from matplotlib import pyplot as plt
    >>> raster_fp = tgp.get_testing_fp('tif_forinterpolation')
    >>> raster = tgp.read_raster(raster_fp)
    >>> data = raster.data.copy()[:, :, 0]
    >>> data_interp = data.copy()
    >>> data_interp[np.isnan(data_interp)] = 999
    >>> data_interp = Interpolation.mean_interpolation(data_interp, no_data_value=999)
    >>> fig, (ax1, ax2) = plt.subplots(1,2)
    >>> ax1.imshow(data)
    >>> ax2.imshow(data_interp)
    >>> plt.show()
    """
    assert len(data.shape) == 2, "data should have onle 2 dimension"
    assert window_size%2==1 , "window_size should be odd number"
    data_interp = __mean_interpolation_single(data, no_data_value=no_data_value, window_size=window_size)
    if loop_to_fill_all and (loop_limit != -1): # with loop_limit
        loop_count = 0
        while (np.sum(data_interp==no_data_value) > 0) and loop_count<loop_limit:
            data_interp = __mean_interpolation_single(data_interp, no_data_value=no_data_value, window_size=window_size)
            loop_count += 1    
    elif loop_to_fill_all and (loop_limit == -1): # without loop_limit
        while np.sum(data_interp==no_data_value) > 0:
            data_interp = __mean_interpolation_single(data_interp, no_data_value=no_data_value, window_size=window_size)
    return data_interp

def gdal_fillnodata(raster, band=0, no_data_value=999, max_distance=100, smoothing_iterations=0):
    """Interpolate values on specific cells (generally nan cell) using 
    `gdal.FillNodata`. To be mentioned, this cannot accept zero in its data
    except its no_data_value is zero.

    Parameters
    ----------
    raster: Raster
        The Raster object you want to fill the no_data_value.
    band: int, optional, default: 0
        The band numnber of Raster object you want to fill the no_data_value 
        which start from zero.
    no_data_value: int or float, optional, default: 999
        The value to be filled with interpolated value. If no_data_value == None, 
        use np.nan as no_data_value.
    max_distance: int, optional, default: 100
        The cells within the max distance from no_data_value location will be 
        calculated to fill the no_data_value.
    smoothing_iterations: int, optional, default: 0
        The maximum limitation on loop.

    Returns
    -------
    data_interp: ndarray. 
        Interpolation result.

    Examples
    -------- 
    >>> import numpy as np
    >>> import TronGisPy as tgp 
    >>> from TronGisPy import Interpolation
    >>> from matplotlib import pyplot as plt
    >>> raster_fp = tgp.get_testing_fp('tif_forinterpolation')
    >>> raster = tgp.read_raster(raster_fp)
    >>> raster.data[np.isnan(raster.data)] = 999
    >>> raster_interp = Interpolation.gdal_fillnodata(raster)
    >>> fig, (ax1, ax2) = plt.subplots(1,2)
    >>> raster.plot(ax=ax1)
    >>> raster_interp.plot(ax=ax2)
    >>> plt.show()
    """    
    # make dst_band
    ds_dst = raster.to_gdal_ds()
    dstband = ds_dst.GetRasterBand(band+1)

    # make maskband
    raster_mask = raster.copy()
    raster_mask.data[raster_mask.data==no_data_value] = 0
    raster_mask.data[raster_mask.data!=0] = 1
    raster_mask.astype(np.bool)
    raster_mask.no_data_value = 0
    ds_mask = raster_mask.to_gdal_ds()
    maskband = ds_mask.GetRasterBand(1)

    gdal.FillNodata(dstband, maskband, max_distance, smoothing_iterations)
    data_interp = tgp.read_gdal_ds(ds_dst)

    ds_dst = None
    ds_mask = None
    return data_interp



import os
import gdal
import numpy as np
import pandas as pd

df_gdal_dtype = [
    [0, "GDT_Unknown", None],
    [1, "GDT_Byte", np.uint8],
    [2, "GDT_UInt16", np.uint16],
    [2, "GDT_UInt16", np.uint16],
    [3, "GDT_Int16", np.int16],
    [4, "GDT_UInt32", np.uint32],
    [5, "GDT_Int32", np.int32],
    [6, "GDT_Float32", np.float32],
    [7, "GDT_Float64", np.float64],
    [8, "GDT_CInt16", None],
    [9, "GDT_CInt32", None],
    [10, "GDT_CFloat32", None],
    [11, "GDT_CFloat64", None],
    [12, "GDT_TypeCount", None],
]
df_gdal_dtype = pd.DataFrame(df_gdal_dtype, columns=['gdaldtype_idx', 'gdaldtype_name', 'npdtype'])

def get_gdaldtype_name(gdaldtype_idx):
    """Get the name of gdal datatype from gdal datatype index.

    Parameters
    ----------
    gdaldtype_idx: int
        Range from 0 to 12 generated from gdal.GDTxxxx e.g. gdal.GDT_Float32.

    Returns
    -------
    gdaldtype_name: str
        Maybe in the following options [GDT_Unknown, GDT_Byte, GDT_UInt16, GDT_Int16, 
        GDT_UInt32, GDT_Int32, GDT_Float32, GDT_Float64, GDT_CInt16, GDT_CInt32, 
        GDT_CFloat32, GDT_CFloat64, GDT_TypeCount]
    """
    gdaldtype_name = df_gdal_dtype.loc[df_gdal_dtype['gdaldtype_idx']==gdaldtype_idx, 'gdaldtype_name'].iloc[0]
    return gdaldtype_name

def gdaldtype_to_npdtype(gdaldtype_idx):
    """Convert the gdal datatype to numpy datatype.

    Parameters
    ----------
    gdaldtype_idx: int
        Range from 0 to 12 generated from gdal.GDTxxxx e.g. gdal.GDT_Float32.

    Returns
    -------
    npdtype: type
        The coresponding numpy datatype for the gdal datatype.
    """
    npdtype = df_gdal_dtype.loc[df_gdal_dtype['gdaldtype_idx']==gdaldtype_idx, 'npdtype'].iloc[0]
    return npdtype

def npdtype_to_gdaldtype(npdtype): 
    """Convert the numpy datatype to gdal datatype.

    Parameters
    ----------
    npdtype: type
        The numpy datatype you want to find the coresponding gdal datatype.

    Returns
    -------
    gdaldtype_idx: int
        Range from 0 to 12 generated from gdal.GDTxxxx e.g. gdal.GDT_Float32.
    """
    warning = os.environ.get('TGPDYPEWARNING')
    warning = str(True) if warning is None else warning
    warning = warning == 'True'

    if npdtype in df_gdal_dtype['npdtype'].tolist():
        return int(df_gdal_dtype.loc[df_gdal_dtype['npdtype']==npdtype, 'gdaldtype_idx'].iloc[0])
    elif np.issubdtype(npdtype, np.bool_):
        return gdal.GDT_Byte
    elif np.issubdtype(npdtype, np.int8): 
        return gdal.GDT_UInt16
    elif np.issubdtype(npdtype, np.uint64):
        if warning: print("cannot find compatible gdaldtype for np.uint64, use gdal.GDT_UInt32 as alternative.")
        return gdal.GDT_UInt32
    elif np.issubdtype(npdtype, np.int64):
        if warning: print("cannot find compatible gdaldtype for np.int64, use gdal.GDT_Int32 as alternative.")
        return gdal.GDT_Int32
    elif np.issubdtype(npdtype, np.signedinteger):
        if warning: print(str(npdtype) + " cannot find compatible gdaldtype.")
        return gdal.GDT_Int16
    elif np.issubdtype(npdtype, np.unsignedinteger):
        if warning: print(str(npdtype) + " cannot find compatible gdaldtype.")
        return gdal.GDT_UInt16
    elif np.issubdtype(npdtype, np.floating):
        if warning: print(str(npdtype) + " cannot find compatible gdaldtype.")
        return gdal.GDT_Float32
    elif np.issubdtype(npdtype, np.generic):
        if warning: print(str(npdtype) + " cannot find compatible gdaldtype.")
        return gdal.GDT_Byte



import os
import setuptools

base_dir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(base_dir, "README.md"), "r") as f:
    long_description = f.read()

setuptools.setup(
    name="TronGisPy",
    version="1.3.8",
    author="Thinktron",
    author_email="jeremywang@thinktronltd.com",
    description="Gis raster data processing tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/thinktron/TronGisPy",
    packages=setuptools.find_packages(),
    # package_data={'PySaga': ['saga_cmd_pkls/*']},
    # package_data={'PySatellite': ['data/*', 'data/*/*']},
    package_data={'TronGisPy': ['data/*', 'data/*/*']},
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
      install_requires=[
          'numba',
          'affine',
          'scikit-learn',
          'descartes',
          'matplotlib',
      ]
)

# 0.0.0
# 0.0.2 modify write_combined_tif to fit the image not well splitted by the right box size
# 0.0.3 add Normalizer
# 0.0.4 change name add CRS and SatelliteIO module
# 0.0.5 add kmeans in Algorithm, clip_image_by_shp & tif_composition in SatelliteIO
# 0.0.6 return no_data_value in get_geo_info
# 0.0.7 add refine_resolution in SatelliteIO
# 0.0.8 add polygonize & rasterize in SatelliteIO
# 0.0.9 add raster_pixel_to_polygon in SatelliteIO, transfer_npidx_to_coord, transfer_coord_to_npidx, transfer_npidx_to_coord_polygon in CRS
# 0.1.0 add get diff testing fp in SateloliteIO
# 0.1.1 add multipolygon in polygonize & zonal function
# 0.1.2 change zonal to default gdal polygonize function
# 0.1.3 add clip_shp_by_shp
# 0.1.4 add LineString in clip_shp_by_shp
# 0.1.5 write output tif can be without projection & geo_transform
# 0.1.6 get_extend return nparray not list

# 0.2.0 change name to TronGisPy
# 0.2.1 no need to give rows, cols, bend in write tif
# 0.2.2 add update_projection
# 0.2.3 cast npdtype to gdaldtype in TypeCast
# 0.2.4 set gdaldtyp in rasterize_layer function
# 0.2.5 change projection, gdadtype and no_data_value to write function in SplittedImage
# 0.2.6 fix SplittedImage.get_geo_attribute coord xy
# 0.2.7 add filter function in SplittedImage.write_splitted_image
# 0.2.8 add self.pad_val in SplittedImage init fun
# 0.2.9 add get_combined_image in SplittedImage
# 0.3.0 step size different from window size is acceptable in SplittedImage 
# 0.3.1 add reproject function in CRS.py
# 0.3.2 get_WKT_from epsg
# 0.3.3 add interpolation function
# 0.3.4 add clipper in Normalizer
# 0.3.5 get_epsg_from_wkt in CRS & remap_tif and reproject in GisIO
# 0.3.6 numba for coords sys transforming between npidx and coords
# 0.3.7 get_extend => get_extent, add param return_poly
# 0.3.8 numba_transfer_group_coord_to_npidx change return type 
# 0.3.9 add all_touched option in rasterize_layer
# 0.4.0 change interpolation function allowed shape to multiple bands image.
# 0.4.1 add transfer_npidx_to_coord
# 0.4.2 add DEMProcessor.py & crop_dem.tif as testing data
# 0.4.3 add function not in CRS
# 0.4.4 change CRS interface: numba_transfer_group_coord_to_npidx => transfer_group_coord_to_npidx
# 0.4.5 remove numba funcation & point to poly
# 0.4.6 transfer_npidx_to_coord_polygon using parallel algorithm
# 0.4.7 add resample alg in refine_resolution
# 0.4.8 fill in nan when reading data in get_nparray of GisIO
# 0.5.0 add ShapeGrid.py to process data in memory, get_nparray => GisIO.get_nparray in test.py
# 0.5.1 add get_extent in ShapeGrid.py 
# 0.5.2 clip_tif_by_bound => clip_tif_by_extent, remove projection fron clip_tif_by_extent params in ShapeGrid.py, return clipped geo_transform in clip_tif_by_extent & clip_tif_by_shp
# 0.5.3 remove projection from refine_resolution params in ShapeGrid.py
# 0.5.4 add geo_transform as return of refine_resolution function in ShapeGrid.py
# 0.5.5 adjust extent input of refine_resolution function in ShapeGrid.py
# 0.5.6 add AeroTriangulation.py and add fast majority interpolation in Interpolation.py
# 0.5.7 add fast mean interpolation in Interpolation.py
# 0.5.8 fast mean interpolation can be float type
# 0.5.9 add no_data_value in ShapeGrid.rasterize_layer
# 0.6.0 activate no_data_value in ShapeGrid.rasterize_layer
# 0.6.1 add loop_limit param in majority_interpolation & mean_interpolation functions 
# 0.6.2 modify geotransform setting in get_geo_attribute function in SplittedImage


# 1.0.0 Raster class for saving raster data in memory
# 1.0.1 update io, Raster, Interpolation, CRS, ShapeGrid documnetation
# 1.0.2 add astype function in Raster.
# 1.0.3 add get_values_by_coords function in Raster.
# 1.0.4 add np.bool in TypeCast, `gdal.FillNodata` in Interpolation
# 1.0.5 add data.copy() when assign data to Raster, and call update_gdaltype_by_npdtype() after assignment
# 1.0.6 add figsize when assign data to Raster, fill no_data when plotting Raster, add note in Interpolation.gdal_fillnodata
# 1.0.7 add gdaltype property in Raster, call paddedimage in `SplittedImage.apply` for efficiency.
# 1.0.8 add change gdaltype to gdaldtype, gdaltype property to gdaldtype_name in Raster, fill_na and fill_no_data in Raster, add fill_na param in `tgp.read_raster`
# 1.0.9 add pixel_size attribute & flipped_gt image ploting in Raster, fix meighbor_majority fill_na test.
# 1.1.0 zonal_stats in ShapeGrid
# 1.1.1 fix bug of flipped_gt image ploting in Raster
# 1.1.2 add cache data for plotting in Raster
# 1.1.3 add hist plotting in Raster
# 1.1.4 clip and norm param in hit and plot of Raster
# 1.1.5 fix nan clipping & fix azimuth (cancel azimuth-=180) in dem_to_hillshade of DEMProcessor
# 1.1.6 fix band_num param in GisIO & band param in DEMProcessor start from 0
# 1.1.7 cache_data_for_plot = None when initializing, fix bands param bug in Raster hist
#       default of flush_cache change to True, fix plot filp error, add resize option
# 1.1.8 Raster geo_transform init with (0,1,0,0,0,-1) if None
# 1.1.9 add get_rasterize_layer_param in ShapeGrid, if single value condition in Raster plot.
# 1.2.0 add documentations 
# 1.2.1 shapegrid clip_raster_with_polygon do not rmove temp dir
# 1.2.2 np.uint64 => gdal.GDT_UInt32 in TypeCast
# 1.2.3 update n_steps in SplittedImage, update items to output Raster in SplittedImage, refine resolution output non-rotated image in ShapeGrid.
# 1.2.4 add idxs_to_be_kept param in write_splitted_images image in SplittedImage.
# 1.2.5 add clip_raster_with_multiple_polygons in ShapeGrid.
# 1.2.6 add clip_min_max and clip_percentage functions and use os.environ to turn off gdaldtype conversion warning.
# 1.2.7 add get_img_pixels_XYZs in AeroTriangulation.
# 1.2.8 fix bug of (L-P) to (p-L) in project_XYZs_to_npidxs, and add return_k param. 
# 1.2.9 fix bug of project_npidxs_to_XYZs
# 1.3.0 fix bug of raster with nan plot using norm clip_percentage
# 1.3.1 use rasterize to clip raster with poly, add ref_raster as input of rasterize layer
# 1.3.2 return AxesImage in Raster.plot function
# 1.3.3 deal with gdal type issue in clip_raster_with_multiple_polygons
# 1.3.4 debug on Raster.get_values_by_coords, ShapeGrid.get_rasterize_layer_params, ShapeGrid.clip_raster_with_polygon
#       move reproject, refine_resolution from ShapeGrid to Raster, add remap_by_ref_raster in Raster.
# 1.3.5 update readme
# 1.3.6 update readme
# 1.3.7 update setup
# 1.3.8 fix reproject in Raster when gdal Warp return None

# python clean_docs_source.py
# sphinx-apidoc --force --separate --module-first -o docs\source .
# cd docs
# make clean
# make html
# cd ..

# rmdir /s/q Y:\R014_Jeremy\temp\TronGisPyDocs
# mkdir Y:\R014_Jeremy\temp\TronGisPyDocs
# XCOPY docs\build\html Y:\R014_Jeremy\temp\TronGisPyDocs /H/S

# python setup.py sdist bdist_wheel
# twine upload dist/TronGisPy-1.3.8*
# scp C:\Users\Thinktron\Projects\TronGisPy\dist\TronGisPy-1.3.8-py3-none-any.whl  jeremy@192.168.0.128:/home/ttl/pypi/TronGisPy-1.3.8-py3-none-any.whl





# scp C:\Users\Thinktron\Projects\TronGisPy\dist\TronGisPy-1.3.8-py3-none-any.whl  jeremy@rd.thinktronltd.com:/home/ttl/pypi/TronGisPy-1.3.8-py3-none-any.whl
# scp C:\Users\TTL_R041\Desktop\Projects\RS2001\LineDetection\TronGisPy\dist\TronGisPy-1.3.8-py3-none-any.whl jeremy@rd.thinktronltd.com:/home/ttl/pypi/TronGisPy-1.3.8-py3-none-any.whl


